"""Chat with a KB."""

import logging
from collections import defaultdict
from dataclasses import dataclass, field
from typing import Any, Callable, ClassVar, Dict, Iterable, Iterator, List, Mapping, Optional

import oaklib.datamodels.obograph as og
from oaklib import BasicOntologyInterface
from oaklib.datamodels.obograph import GraphDocument
from oaklib.datamodels.search import SearchConfiguration
from oaklib.datamodels.vocabulary import IS_A
from oaklib.interfaces import OboGraphInterface, SearchInterface
from oaklib.types import CURIE
from oaklib.utilities.iterator_utils import chunk

from curategpt import DBAdapter
from curategpt.formatters.format_utils import camelify
from curategpt.wrappers.base_wrapper import BaseWrapper
from curategpt.wrappers.ontology.ontology import OntologyClass, Relationship

logger = logging.getLogger(__name__)


@dataclass
class OntologyWrapper(BaseWrapper):
    """
    A wrapper to pull from ontologies using OAK.

    This wrapper can be used either with static sources (e.g. an ontology file)
    or dynamic (pubmed).

    The data model is a simple JSON/Dict structure
    """

    name: ClassVar[str] = "oaklib"

    oak_adapter: BasicOntologyInterface = None

    id_to_shorthand: Mapping[CURIE, CURIE] = None
    shorthand_to_id: Mapping[CURIE, CURIE] = None
    _objects_by_curie: Mapping[CURIE, OntologyClass] = None
    _objects_by_shorthand: Mapping[str, OntologyClass] = None

    default_max_search_results: int = 500
    fetch_definitions: bool = field(default=True)
    fetch_relationships: bool = field(default=True)
    relationships_as_fields: bool = field(default=False)

    branches: List[str] = None

    def __post_init__(self):
        """Initialize."""
        self.id_to_shorthand = {}

    def external_search(self, text: str, expand: bool = True, limit: int = None, **kwargs) -> List:
        adapter = self.oak_adapter
        if not isinstance(adapter, SearchInterface):
            raise ValueError(f"OAK adapter {self.oak_adapter} does not support search")
        cfg = SearchConfiguration(is_partial=True)
        chunk_iter = chunk(
            adapter.basic_search(text, cfg), limit or self.default_max_search_results
        )
        for object_ids in chunk_iter:
            print("object_ids", object_ids)
            return list(self.objects(object_ids=object_ids))

    def objects(
        self, collection: str = None, object_ids: Optional[Iterable[str]] = None, **kwargs
    ) -> Iterator[Dict]:
        """
        Yield all objects in the view.

        :return:
        """
        adapter = self.oak_adapter
        entities = list(adapter.entities())

        if self.branches:
            if not isinstance(adapter, OboGraphInterface):
                raise ValueError(f"OAK adapter {self.oak_adapter} does not support branches")
            selected_ids = []
            for branch in self.branches:
                selected_ids.extend(list(adapter.descendants([branch], predicates=[IS_A])))
            selected_ids = list(set(selected_ids))
        elif object_ids:
            selected_ids = list(object_ids)
        else:
            selected_ids = list(entities)
        logger.info(f"Found {len(selected_ids)} selected ids")
        # need to fetch ALL labels in store, even if not selected,
        # as may be used in references
        labels = {e: lbl for e, lbl in adapter.labels(entities, allow_none=False)}
        logger.info(f"Found {len(labels)} labels")
        definitions = {}
        if self.fetch_definitions:
            for chunked_entities in chunk(selected_ids, 100):
                for id, defn, _ in adapter.definitions(chunked_entities):
                    definitions[id] = defn
        relationships = defaultdict(list)
        if self.fetch_relationships:
            for sub, pred, obj in adapter.relationships():
                if sub in entities:
                    relationships[sub].append((pred, obj))
        self.id_to_shorthand = {}
        self.shorthand_to_id = {}
        for id, lbl in labels.items():
            shorthand = camelify(lbl)
            if shorthand in self.shorthand_to_id:
                shorthand = f"{shorthand}_{id}"
            if shorthand in self.shorthand_to_id:
                continue
            self.id_to_shorthand[id] = shorthand
            self.shorthand_to_id[shorthand] = id
        self._objects_by_curie = {}
        self._objects_by_shorthand = {}
        for id, shorthand in self.id_to_shorthand.items():
            if id not in selected_ids:
                continue
            obj = OntologyClass(
                id=shorthand,
                label=labels[id],
                original_id=id,
            )
            for pred, tgt in relationships.get(id, []):
                k = self._as_shorthand(pred)
                k = k.replace("rdfs:", "")
                if self.relationships_as_fields:
                    if not hasattr(obj, k):
                        setattr(obj, k, [])
                    getattr(obj, k).append(self._as_shorthand(tgt))
                else:
                    if not obj.relationships:
                        obj.relationships = []
                    obj.relationships.append(
                        Relationship(predicate=k, target=self._as_shorthand(tgt))
                    )
            if id in definitions:
                obj.definition = definitions[id]
            self._objects_by_curie[id] = obj
        # for id, alias in adapter.alias_relationships():
        if isinstance(adapter, OboGraphInterface):
            for ldef in adapter.logical_definitions():
                shorthand = self._as_shorthand(ldef.definedClassId)
                obj = self._objects_by_curie.get(ldef.definedClassId, None)
                if obj is None:
                    continue
                obj.logical_definition = [
                    Relationship(predicate=IS_A, target=self._as_shorthand(g))
                    for g in ldef.genusIds
                ] + [
                    Relationship(
                        predicate=self._as_shorthand(r.propertyId),
                        target=self._as_shorthand(r.fillerId),
                    )
                    for r in ldef.restrictions
                ]
        for obj in self._objects_by_curie.values():
            yield obj.dict()

    def as_object(self, curie: CURIE) -> Optional[OntologyClass]:
        if not self._objects_by_curie:
            list(self.objects())
        return self._objects_by_curie.get(curie, None)

    def _as_shorthand(self, curie: CURIE) -> CURIE:
        return self.id_to_shorthand.get(curie, curie)

    @property
    def text_field(self) -> Callable:
        """
        Returns a function that returns the embeddable text field for an object.

        :return:
        """
        # TODO: decouple from OntologyClass
        return lambda obj: obj.label if isinstance(obj, OntologyClass) else obj.get("label")

    def retrieve_shorthand_to_id_from_store(self, store: DBAdapter) -> Mapping[str, str]:
        if not self.shorthand_to_id:
            self.shorthand_to_id = {}
            for obj, _, __ in store.find({}):
                self.shorthand_to_id[obj["id"]] = obj["original_id"]
        return self.shorthand_to_id

    def unwrap_object(self, obj: Dict[str, Any], store: DBAdapter, **kwargs) -> GraphDocument:
        return self.unwrap_objects([obj], store, **kwargs)

    def unwrap_objects(
        self, objs: Iterable[Dict[str, Any]], store: DBAdapter, drop_dangling=False, **kwargs
    ) -> og.GraphDocument:
        """
        Convert an object from the store to the view representation.

        reverse transform of `as_object`

        :param object:
        :param kwargs:
        :return:
        """
        m = self.retrieve_shorthand_to_id_from_store(store)
        graph = og.Graph(id="tmp", nodes=[], edges=[])
        for obj in objs:
            if not obj:
                logger.warning(f"Skipping empty object {obj}")
                continue
            id = obj.get("original_id", None)
            if not id:
                logger.warning(f"Skipping empty id in {obj}")
                continue
            node = og.Node(
                id=id,
                lbl=obj.get("label", None),
                type="CLASS",
            )
            meta = {}
            defn = obj.get("definition", None)
            if defn:
                meta["definition"] = {"val": defn}
            node.meta = meta
            graph.nodes.append(node)
            for rel in obj.get("relationships", []):
                tgt = rel.get("target", None)
                if not tgt:
                    logger.warning(f"Skipping empty target in {rel}")
                    continue
                if isinstance(tgt, list):
                    logger.warning(f"Unexpected list: {tgt}")
                    tgt = tgt[0]
                if tgt in m:
                    tgt = m[tgt]
                else:
                    if drop_dangling:
                        logger.warning(f"Could not find {tgt}")
                        continue
                    else:
                        tgt = tgt.replace(":", "_")
                pred = rel["predicate"]
                if pred in m:
                    pred = m[pred]
                else:
                    logger.warning(f"Could not find {pred}")
                if pred == "subClassOf":
                    pred = "is_a"
                edge = og.Edge(
                    sub=id,
                    obj=tgt,
                    pred=pred,
                )
                graph.edges.append(edge)
        return og.GraphDocument(graphs=[graph])



"""Annotation (Concept Recognition) in texts."""

import logging
from dataclasses import dataclass
from enum import Enum
from typing import Dict, List, Optional, Tuple

from pydantic import BaseModel, ConfigDict

from curategpt.agents.base_agent import BaseAgent

logger = logging.getLogger(__name__)


CONCEPT = Tuple[str, str]


class Span(BaseModel):
    """An individual span of text containing a single concept."""

    model_config = ConfigDict(protected_namespaces=())

    text: str

    start: Optional[int] = None
    end: Optional[int] = None

    concept_id: str = None
    """Concept ID."""

    concept_label: Optional[str] = None
    """Concept label."""

    is_suspect: Optional[bool] = False
    """Potential hallucination due to ID/label mismatch."""


class GroundingResult(BaseModel):
    """Result of grounding text."""

    model_config = ConfigDict(protected_namespaces=())

    input_text: str
    """Text that is supplied for grounding, assumed to contain a single context."""

    spans: Optional[List[Span]] = []
    """Ordered list of candidate spans."""

    score: Optional[float] = None
    """Score/confidence, from zero to one."""


class AnnotationMethod(str, Enum):
    """Strategy or algorithm used for CR."""

    INLINE = "inline"
    """LLM creates an annotated document"""

    CONCEPT_LIST = "concept_list"
    """LLM creates a list of concepts"""

    TWO_PASS = "two_pass"
    """LLM annotates a document using NER and then grounds the concepts"""


class AnnotatedText(BaseModel):
    """In input text annotated with concept instances."""

    model_config = ConfigDict(protected_namespaces=())

    input_text: str
    """Text that is supplied for annotation."""

    concepts: Optional[Dict[str, str]] = {}
    """Dictionary of concepts found in the text. TODO: change to list of spans."""

    annotated_text: Optional[str] = None
    """Text with concepts annotated (not all methods produce this)."""

    spans: Optional[List[Span]] = []

    summary: Optional[str] = None
    """Summary of the results."""

    prompt: Optional[str] = None
    """Prompt used to generate the annotated text."""


GROUND_PROMPT = """
Your role is to assign a concept ID that best matches the supplied text, using
the supplied list of candidate concepts.
Return as a string "CONCEPT NAME // CONCEPT ID".
Only return a result if the input text represents the same or equivalent
concept, in the provided context.
If there is no match, return an empty string.
"""


xxx_GROUND_PROMPT = """
Your role is to assign a concept ID that best matches the supplied text, using
the supplied list of candidate concepts.
Return results as a CSV of ID,label,score triples, where the ID is the concept.
Only use concept IDs from the supplied list of candidate concepts.
Only return a row if the concept ID is a match for the input text
If there is no match, return an empty string.
"""

MENTION_PROMPT = """
Your role is to list all instances of the supplied candidate concepts in the supplied text.
Return the concept instances as a CSV of ID,label,text pairs, where the ID
is the concept ID, label is the concept label, and text is the mention of the
concept in the text.
The concept ID and label should come only from the list of candidate concepts supplied to you.
Only include a row if the meaning of the text section is that same as the concept.
If there are no instances of a concept in the text, return an empty string.
Do not include additional verbiage.
"""

ANNOTATE_PROMPT = """
Your role is to annotate the supplied text with selected concepts.
return the original text with each conceptID in square brackets.
After the occurrence of that concept.
You can use synonyms. For example, if the concept list contains
zucchini // DB:12345
Then for the text 'I love courgettes!' you should return
'I love [courgettes DB:12345]!'
Always try and match the longest span.
The concept ID should come only from the list of candidate concepts supplied to you.
"""


import re


def parse_annotations(text, marker_char: str = None) -> List[CONCEPT]:
    """
    Parse annotations from text.

    >>> text = ("A minimum diagnostic criterion is the combination of either the [skin tumours] or multiple "
    ...        "[odontogenic keratocysts HP:0010603] of the jaw plus a positive [family history HP:0032316] "
    ...        "for this disorder, [bifid ribs HP:0000923], lamellar [calcification of falx cerebri HP:0005462] "
    ...        "or any one of the skeletal abnormalities typical of this syndrome")
    >>> for ann in parse_annotations(text):
    ...    print(ann)
    ('skin tumours', None)
    ('odontogenic keratocysts', 'HP:0010603')
    ('family history', 'HP:0032316')
    ('bifid ribs', 'HP:0000923')
    ('calcification of falx cerebri', 'HP:0005462')

    For texts with marker characters:

    >>> text = "for this disorder, [bifid ribs | HP:0000923], lamellar [calcification of falx cerebri | HP:0005462] "
    >>> for ann in parse_annotations(text, "|"):
    ...    print(ann)
    ('bifid ribs', 'HP:0000923')
    ('calcification of falx cerebri', 'HP:0005462')

    :param text:
    :return:
    """
    # First Pass: Extract text within [ ... ]
    pattern1 = r"\[([^\]]+)\]"
    matches = re.findall(pattern1, text)

    # Second Pass: Parse the last token of each match
    annotations = []
    for match in matches:
        # Split the match into words and check the last token
        if marker_char:
            toks = match.split(marker_char)
            if len(toks) > 1:
                annotation = " ".join(toks[:-1]).strip()
                id = toks[-1].strip()
            else:
                annotation = match
                id = None
        else:
            words = match.split()
            if len(words) > 1 and ":" in words[-1]:
                annotation = " ".join(words[:-1])
                id = words[-1]
            else:
                annotation = match
                id = None

        annotations.append((annotation, id))

    return annotations


def parse_spans(text: str, concept_dict: Dict[str, str] = None) -> List[Span]:
    spans = []
    for line in text.split("\n"):
        logger.debug(f"Line: {line}")
        row = line.split(",")
        if len(row) < 2:
            logger.debug(f"Skipping line: {line}")
            continue
        concept_id = row[0].strip('"')
        if concept_id == "ID":
            continue
        if " " in concept_id:
            continue
        concept_label = row[1].strip('"')
        mention_text = ",".join(row[2:])
        verified_concept_label = concept_dict.get(concept_id, None)
        spans.append(
            Span(
                text=mention_text,
                concept_id=concept_id,
                concept_label=verified_concept_label,
                is_suspect=verified_concept_label != concept_label,
            )
        )
    return spans


@dataclass
class ConceptRecognitionAgent(BaseAgent):
    identifier_field: str = None
    """Field to use as identifier for objects."""

    label_field: str = None
    """Field to use as label for objects."""

    split_input_text: bool = None

    relevance_factor: float = 0.8
    """Relevance factor for diversifying search results using MMR."""

    prefixes: List[str] = None
    """List of prefixes to use for concept IDs."""

    def ground_concept(
        self,
        text: str,
        collection: str = None,
        categories: Optional[List[str]] = None,
        include_category_in_search=True,
        context: str = None,
        **kwargs,
    ) -> GroundingResult:
        system_prompt = GROUND_PROMPT
        query = text
        if include_category_in_search and categories:
            query += " Categories: " + ", ".join(categories)
        concept_pairs, concept_prompt = self._label_id_pairs_prompt_section(
            query, collection, **kwargs
        )
        concept_dict = {c[0]: c[1] for c in concept_pairs}
        system_prompt += concept_prompt
        model = self.extractor.model
        logger.debug(f"Prompting with: {text}")
        if context:
            prompt_text = f"The overall context for this is the sentence '{context}'.\n\n"
            prompt_text += f"Concept to ground: {text}"
        else:
            prompt_text = f"Concept to ground: {text}"
        response = model.prompt(prompt_text, system=system_prompt)
        logger.debug(f"Response: {response.text()}")
        lines = response.text().split("\n")
        spans = []
        for line in lines:
            if "//" in line:
                toks = line.split("//")
                if len(toks) > 2:
                    logger.warning(f"Multiple concepts in one line: {line}")
                concept_label, concept_id = toks[0], toks[1]
                concept_id = concept_id.strip()
                if " " in concept_id:
                    continue
                provided_concept_label = concept_label.strip()
                if concept_id in concept_dict:
                    concept_label = concept_dict[concept_id]
                else:
                    concept_label = None
                span = Span(
                    text=text,
                    concept_id=concept_id,
                    concept_label=concept_label,
                    is_suspect=provided_concept_label != concept_label,
                )
                spans.append(span)
        # spans = parse_spans(response.text(), concept_dict)
        ann = GroundingResult(input_text=text, annotated_text=response.text(), spans=spans)
        return ann

    def annotate(
        self,
        text: str,
        collection: str = None,
        method=AnnotationMethod.INLINE,
        **kwargs,
    ) -> AnnotatedText:
        if method == AnnotationMethod.INLINE:
            return self.annotate_inline(text, collection, **kwargs)
        elif method == AnnotationMethod.CONCEPT_LIST:
            return self.annotate_concept_list(text, collection, **kwargs)
        elif method == AnnotationMethod.TWO_PASS:
            return self.annotate_two_pass(text, collection, **kwargs)
        else:
            raise ValueError(f"Unknown annotation method {method}")

    def annotate_two_pass(
        self,
        text: str,
        collection: str = None,
        categories: List[str] = None,
        **kwargs,
    ) -> AnnotatedText:
        if not categories:
            categories = ["NamedEntity"]
        system_prompt = "Your job is to parse the supplied text, identifying instances of concepts "
        if len(categories) == 1:
            system_prompt += f" that represent some kind of {categories[0]}. "
            system_prompt += (
                "Mark up the concepts in square brackets, "
                "preserving the original text inside the brackets. "
            )
        else:
            system_prompt += " that represent one of the following categories: "
            system_prompt += ", ".join(categories)
            system_prompt += (
                "Mark up the concepts in square brackets, with the category after the pipe symbol, "
            )
            system_prompt += "Using the syntax [ORIGINAL TEXT | CATEGORY]."
        logger.debug(f"Prompting with: {text}")
        model = self.extractor.model
        response = model.prompt(text, system=system_prompt)
        marked_up_text = response.text()
        anns = parse_annotations(marked_up_text, "|")
        spans = []
        for term, category in anns:
            concept = self.ground_concept(
                term,
                collection,
                categories=[category] if category else None,
                context=text,
                **kwargs,
            )
            if not concept.spans:
                logger.debug(f"Unable to ground concept {term} in category {category}")
                continue
            main_span = concept.spans[0]
            spans.append(
                Span(
                    text=term,
                    concept_id=main_span.concept_id,
                    concept_label=main_span.concept_label,
                )
            )
        return AnnotatedText(
            input_text=text,
            annotated_text=marked_up_text,
            spans=spans,
        )

    def annotate_inline(
        self,
        text: str,
        collection: str = None,
        categories: List[str] = None,
        **kwargs,
    ) -> AnnotatedText:
        system_prompt = ANNOTATE_PROMPT
        concept_pairs, concepts_prompt = self._label_id_pairs_prompt_section(
            text, collection, **kwargs
        )
        concept_dict = {c[0]: c[1] for c in concept_pairs}
        system_prompt += concepts_prompt
        model = self.extractor.model
        logger.debug(f"Prompting with: {text}")
        response = model.prompt(text, system=system_prompt)
        anns = parse_annotations(response.text())

        logger.info(f"Anns: {anns}")
        spans = [
            Span(text=ann[0], concept_id=ann[1], concept_label=concept_dict.get(ann[1], None))
            for ann in anns
        ]
        return AnnotatedText(input_text=text, spans=spans, annotated_text=response.text())

    def annotate_concept_list(
        self,
        text: str,
        collection: str = None,
        categories: List[str] = None,
        **kwargs,
    ) -> AnnotatedText:
        system_prompt = MENTION_PROMPT
        concept_pairs, concepts_prompt = self._label_id_pairs_prompt_section(
            text, collection, **kwargs
        )
        concept_dict = {c[0]: c[1] for c in concept_pairs}
        system_prompt += concepts_prompt
        model = self.extractor.model
        logger.debug(f"Prompting with: {text}")
        response = model.prompt(text, system=system_prompt)
        spans = parse_spans(response.text(), concept_dict)
        return AnnotatedText(
            input_text=text, summary=response.text(), spans=spans, prompt=system_prompt
        )

    def _label_id_pairs_prompt_section(
        self,
        text: str,
        collection: str,
        prolog: str = None,
        relevance_factor: float = None,
        **kwargs,
    ) -> Tuple[List[CONCEPT], str]:
        prompt = prolog
        if not prompt:
            prompt = "Here are the candidate concepts, as label // ConceptID pairs:\n"
        id_field = self.identifier_field
        label_field = self.label_field
        if not id_field:
            id_field = "id"
        if not label_field:
            label_field = "label"
        if relevance_factor is None:
            relevance_factor = self.relevance_factor
        logger.debug(f"System prompt = {prompt}")
        concept_pairs = []
        for obj, _, _obj_meta in self.knowledge_source.search(
            text,
            relevance_factor=relevance_factor,
            collection=collection,
            **kwargs,
        ):
            id, label = obj.get(id_field, None), obj.get(label_field, None)
            if self.prefixes:
                if not any(id.startswith(prefix + ":") for prefix in self.prefixes):
                    continue
            if not id:
                raise ValueError(f"Object {obj} has no ID field {id_field}")
            if not label:
                raise ValueError(f"Object {obj} has no label field {label_field}")
            prompt += f"{label} // {id}   \n"
            concept_pairs.append((id, label))
        return concept_pairs, prompt



"""Basic Extractor that is purely example driven."""

import json
import logging
import re
from copy import copy
from dataclasses import dataclass
from typing import List

import yaml
from pydantic import ConfigDict

from curategpt.formatters.format_utils import remove_formatting

from ..utils.tokens import estimate_num_tokens, max_tokens_by_model
from .extractor import AnnotatedObject, Extractor

logger = logging.getLogger(__name__)


@dataclass
class BasicExtractor(Extractor):
    """
    Extractor that is purely example driven.
    """

    model_config = ConfigDict(protected_namespaces=())

    serialization_format: str = "json"
    model_name: str = "gpt-4o"

    def extract(
        self,
        text: str,
        target_class: str,
        examples: List[AnnotatedObject] = None,
        background_text: str = None,
        rules: List[str] = None,
        min_examples=1,
        **kwargs,
    ) -> AnnotatedObject:
        logger.debug(f"Basic extractor: {text}, {len(examples)} examples")
        examples = copy(examples)
        while True:
            prompt = ""
            if background_text:
                prompt += f"{background_text}\n\n"
            prompt += f"Extract a {target_class} object from text in {self.serialization_format} format.\n\n"
            if rules:
                prompt += "Rules:\n\n"
                for rule in rules:
                    prompt += f"- {rule}\n"
                prompt += "\n"
                prompt += "---\n"
            prompt += "Examples:\n\n"
            for example in examples:
                if example.text:
                    prompt += f"##\nText: {example.text}\n"
                prompt += f"Response: {self.serialize(example)}\n"
            prompt += f"\n##\nText: {text}\n\n"
            prompt += "Response: "
            estimated_length = estimate_num_tokens([prompt])
            if estimated_length + 300 < max_tokens_by_model(self.model.model_id):
                break
            else:
                # remove least relevant
                logger.debug(f"Removing least relevant of {len(examples)}: {examples[-1]}")
                examples.pop()
                if len(examples) < min_examples:
                    raise ValueError(
                        f"Prompt too long, need at least {min_examples} examples: {prompt}."
                    )
        model = self.model
        logger.info(f"Prompt: {prompt}")
        response = model.prompt(prompt)
        ao = self.deserialize(response.text())
        ao.annotations["prompt"] = prompt
        return ao

    def serialize(self, ao: AnnotatedObject) -> str:
        if self.serialization_format == "yaml":
            return yaml.dump(ao.object, sort_keys=False)
        else:
            return json.dumps(ao.object)

    def deserialize(self, text: str, format=None, **kwargs) -> AnnotatedObject:
        if format is None:
            format = self.serialization_format
        if format == "yaml":
            return self.deserialize_yaml(text, **kwargs)
        logger.debug(f"Parsing {text}")
        text = remove_formatting(text=text, expect_format="json")
        try:
            obj = json.loads(text)
            if isinstance(obj, str):
                if self.raise_error_if_unparsable:
                    raise ValueError(f"Could not parse {text}")
                else:
                    obj = {}
            return AnnotatedObject(object=obj)
        except Exception as e:
            match = re.search(r"\{.*\}", text)
            if match:
                if match.group() != text:
                    return self.deserialize(match.group())
            if self.raise_error_if_unparsable:
                raise e
            else:
                logger.warning(f"Could not parse {text}")
                return AnnotatedObject(object={})

    def deserialize_yaml(self, text: str, multiple=False) -> AnnotatedObject:
        logger.debug(f"Parsing YAML: {text}")
        text = remove_formatting(text=text, expect_format="yaml")
        try:
            if multiple:
                obj = yaml.safe_load_all(text)
            else:
                obj = yaml.safe_load(text)
            if isinstance(obj, str):
                if self.raise_error_if_unparsable:
                    raise ValueError(f"Could not parse {text}")
                else:
                    obj = {}
            return AnnotatedObject(object=obj)
        except Exception as e:
            logger.warning(f"Could not parse {text} due to {e}")
            return AnnotatedObject(object={})



"""Chat with a KB."""

import logging
from copy import deepcopy
from csv import DictReader
from dataclasses import dataclass
from functools import lru_cache
from typing import ClassVar, Dict, Iterable, Iterator, Optional, TextIO

import requests
from oaklib import BasicOntologyInterface, get_adapter

from curategpt.wrappers import BaseWrapper
from curategpt.wrappers.literature.pubmed_wrapper import PubmedWrapper

logger = logging.getLogger(__name__)

MAP = {
    "disease_id": "disease",
    "disease_name": "disease_label",
    "hpo_id": "phenotype",
    "maxo_id": "treatment",
    "maxo_label": "treatment_label",
}


def filter_header(row) -> bool:
    return row[0] != "#"


def stream_filtered_lines(response):
    """Generator to yield non-comment lines from a streaming response."""
    buffer = ""
    for chunk in response.iter_content(chunk_size=8192):
        # Decode the chunk of bytes to string
        buffer += chunk.decode("utf-8")
        while "\n" in buffer:
            line, buffer = buffer.split("\n", 1)
            if not line.strip().startswith("#"):
                yield line


@lru_cache
def hpo_ont_adapter():
    return get_adapter("sqlite:obo:hp")


@lru_cache
def maxo_ont_adapter():
    return get_adapter("sqlite:obo:maxp")


@lru_cache
def term_label(identifier: str) -> str:
    return hpo_ont_adapter().label(identifier)


@dataclass
class MAXOAWrapper(BaseWrapper):
    """
    A wrapper over HPOA

    """

    name: ClassVar[str] = "maxoa"

    source_url: ClassVar[str] = (
        "https://raw.githubusercontent.com/monarch-initiative/"
        "maxo-annotations/master/annotations/maxo-annotations.tsv"
    )

    expand_publications: bool = True

    _label_adapter: BasicOntologyInterface = None

    default_object_type = "DiseaseTreatmentAssociation"

    pubmed_wrapper: PubmedWrapper = None

    group_by_publication: bool = True

    def __post_init__(self):
        from curategpt.wrappers.literature.pubmed_wrapper import PubmedWrapper

        self.pubmed_wrapper = PubmedWrapper()
        # reuse hpoa cache
        self.pubmed_wrapper.set_cache("hpoa_pubmed_cache")

    def objects(
        self,
        collection: str = None,
        object_ids: Optional[Iterable[str]] = None,
        source_locator=None,
        **kwargs,
    ) -> Iterator[Dict]:
        # open a file handle from a URL using requests
        if source_locator is None:
            source_locator = self.source_locator
        if source_locator is None:
            source_locator = self.source_url
        if source_locator.startswith("http"):
            logger.info(f"Fetching {source_locator}")
            with requests.get(source_locator, stream=True) as response:
                response.raise_for_status()  # Raise an error for failed requests
                reader = DictReader(stream_filtered_lines(response), delimiter="\t")
                yield from self.objects_from_rows(reader)
        else:
            with open(source_locator) as file:
                reader = DictReader(filter(filter_header, file), delimiter="\t")
                yield from self.objects_from_rows(reader)

    def objects_from_rows(self, rows: Iterable[Dict]) -> Iterator[Dict]:
        by_pub = {}
        for row in rows:
            row = {MAP.get(k, k): v for k, v in row.items()}
            row["phenotype_label"] = term_label(row["phenotype"])
            refs = [ref for ref in row["citation"].split(";") if ref != "PMID:UNKNOWN"]
            if self.expand_publications:
                pmids = [ref for ref in refs if ref.startswith("PMID")]
                logger.debug(f"Expanding {refs}, pmids={pmids}")
                if pmids:
                    pubs = self.pubmed_wrapper.objects_by_ids(pmids)
                    row["publications"] = pubs
                    for pub in row["publications"]:
                        pub_id = pub["id"]
                        if pub_id not in by_pub:
                            by_pub[pub_id] = deepcopy(pub)
                            by_pub[pub_id]["associations"] = []
                            logger.info(f"Adding {pub_id} to by_pub, {len(by_pub)} total")
                        by_pub[pub_id]["associations"].append(
                            {k: v for k, v in row.items() if k != "publications"}
                        )
            if not self.group_by_publication:
                yield row
        if self.group_by_publication:
            for pub in by_pub.values():
                yield pub

    def objects_from_file(self, file: TextIO) -> Iterator[Dict]:
        rows = DictReader(filter(filter_header, file), delimiter="\t")
        yield from self.objects_from_rows(rows)



"""Utilities for streamlit app."""

from typing import Dict, List, Optional

import yaml

from curategpt.conf import EXAMPLES_DIR

HELP_CASES = EXAMPLES_DIR / "bio.yaml"


def get_case_collection():
    with open(HELP_CASES) as stream:
        return yaml.safe_load(stream)


def get_applicable_examples(collection: Optional[str], mode: str, relax=True) -> List[Dict]:
    """
    Get applicable examples for a given collection and mode.

    :param collection:
    :param mode:
    :return:
    """
    if mode:
        mode = mode.upper()
    examples = []
    cases = get_case_collection()["cases"]
    for case in cases:
        if mode and case["mode"] != mode:
            continue
        if collection and case["source"] not in collection:
            # TODO: less hacky check
            continue
        case = {k: v for k, v in case.items() if k not in ["domains", "answers"]}
        examples.append(case)
    if not examples and relax and collection:
        # If no examples are found, try to relax the collection
        return get_applicable_examples(None, mode, relax=False)
    return examples



"""Chat with a Google Drive."""

import logging
from dataclasses import dataclass
from typing import Any, ClassVar, Dict, Iterable, Iterator, Optional

import gspread

from curategpt.wrappers.base_wrapper import BaseWrapper

logger = logging.getLogger(__name__)


# Authenticate and build the service
SCOPES = ["https://www.googleapis.com/auth/drive.readonly"]


@dataclass
class GSpreadWrapper(BaseWrapper):
    """
    A wrapper to provide a search facade over gspread layer onto google sheets.

    Currently static, but could be made dynamic in the future.
    """

    name: ClassVar[str] = "gspread"

    sheet_name: str = None
    worksheet_name: str = None

    service: Any = None
    gc = None

    def __post_init__(self):
        if self.source_locator:
            if "/" in self.source_locator:
                self.sheet_name, self.worksheet_name = self.source_locator.split("/")
            else:
                self.sheet_name = self.source_locator
        self.gc = gspread.service_account()

    def objects(
        self, collection: str = None, object_ids: Optional[Iterable[str]] = None, **kwargs
    ) -> Iterator[Dict]:
        logger.info(f"Loading {self.sheet_name}/{self.worksheet_name}")
        wks = self.gc.open(str(self.sheet_name)).worksheet(self.worksheet_name)
        yield from wks.get_all_records()



import logging
from dataclasses import dataclass, field
from enum import Enum
from typing import Dict, List, Optional, Union

import yaml

from curategpt import BasicExtractor
from curategpt.agents.base_agent import BaseAgent
from curategpt.agents.chat_agent import ChatAgent, ChatResponse
from curategpt.formatters.format_utils import object_as_yaml
from curategpt.utils.tokens import estimate_num_tokens, max_tokens_by_model
from curategpt.wrappers import BaseWrapper

logger = logging.getLogger(__name__)


class EvidenceUpdatePolicyEnum(str, Enum):
    skip = "skip"
    append = "append"
    replace = "replace"


@dataclass
class EvidenceAgent(BaseAgent):
    """
    An agent to find evidence for an object by querying a reference source.

    The evidence agent is able to find (supporting and refuting) evidence for any of
    the following:

    - A simple statement in natural language
    - A simple structured dictionary object of key-value pairs
    - A complex structured dictionary object with nested key-value pairs

    The default source used is Pubmed, using :class:`PubmedWrapper` via :class:`ChatAgent`
    """

    chat_agent: Union[ChatAgent, BaseWrapper] = None

    evidence_update_policy: EvidenceUpdatePolicyEnum = field(default=EvidenceUpdatePolicyEnum.skip)

    def find_evidence(self, obj: Union[str, Dict]) -> ChatResponse:
        obj_as_str = obj if isinstance(obj, str) else object_as_yaml(obj)
        text = f"Find evidence in the given references to support an object: {obj_as_str}"
        return self.chat_agent.chat(text)

    def find_evidence_simple(self, query: str, limit: int = 10, **kwargs) -> Optional[List[Dict]]:
        logger.info(f"Finding evidence for: {query}")
        extractor = self.chat_agent.extractor
        if extractor is None:
            if isinstance(self.knowledge_source, BaseWrapper):
                extractor = self.knowledge_source.extractor
            if extractor is None:
                logger.warning("No extractor found. Using default.")
                extractor = BasicExtractor()
        chat_agent = self.chat_agent
        if isinstance(self.chat_agent, BaseWrapper):
            chat_agent = ChatAgent(
                knowledge_source=chat_agent,
                extractor=extractor,
            )
        collection = None
        # collection = self.chat_agent.knowledge_source_collection
        if collection is None:
            collection = self.knowledge_source_collection
        kwargs["collection"] = collection
        kb_results = list(
            chat_agent.knowledge_source.search(
                query, relevance_factor=chat_agent.relevance_factor, limit=limit, **kwargs
            )
        )
        # TODO: DRY
        while True:
            i = 0
            references = {}
            texts = []
            current_length = 0
            for obj, _, _obj_meta in kb_results:
                i += 1
                obj_text = yaml.dump({k: v for k, v in obj.items() if v}, sort_keys=False)
                references[str(i)] = obj_text
                texts.append(f"## Reference\n{obj_text}")
                current_length += len(obj_text)
            model = extractor.model
            logger.info(f"Using model: {model.model_id}")
            prompt = "---\nBackground literature:\n"
            prompt += "\n".join(texts)
            prompt += f"---\nHere is the Statement: {query}.\n"
            logger.debug(f"Candidate Prompt: {prompt}")
            estimated_length = estimate_num_tokens([prompt])
            logger.debug(
                f"Max tokens {extractor.model.model_id}: {max_tokens_by_model(extractor.model.model_id)}"
            )
            # TODO: use a more precise estimate of the length
            if estimated_length + 300 < max_tokens_by_model(extractor.model.model_id):
                break
            else:
                # remove least relevant
                logger.debug(f"Removing least relevant of {len(kb_results)}: {kb_results[-1]}")
                if not kb_results:
                    raise ValueError(f"Prompt too long: {prompt}.")
                kb_results.pop()

        logger.debug(f"Prompt: {prompt}")

        response = model.prompt(
            prompt,
            system="""
        You are a scientist assistant. Given a statement in subject-predicate-object form, your job is
        to look at the literature I provide you, and return an object that states whether the literature
        supports or refutes the statement.

        Return in YAML:

        ```
        - reference: IDENTIFIER  ## e.g PMID:123456
          supports: "SUPPORT" | "REFUTE" | "NO_EVIDENCE" | "PARTIAL" | "WRONG_STATEMENT"
          snippet: "EXCERPT"
          explanation: "<explanatory text, if necessary>"
        - ...
        ```

        Never put a space between the pubmed prefix and local ID. The correct way to write is PMID:123456.

        Do not include an entry if you cannot find it in the literature provided.

        Pay close attention to the predicate. For example, the following statement:

            Subject: The Sun Predicate: orbits Value: name: Mercury

        is false, because the subject and value are inverted. Use WRONG_STATEMENT in this case.

        Similarly, the following statement:

            Subject: Mercury Predicate: friend-of Value: name: The Sun

        Is false because the predicate is not correct for these two entities.
        Use WRONG_STATEMENT in this case.

        Additionally, if the statement is:

            Subject: Ebola Predicate: causes Value: {name: Bleeding, severity: Mild }

        And the literature says that Ebola causes SEVERE bleeding.
        Use WRONG_STATEMENT in this case.
        """,
        )
        response_text = response.text()
        if "```" in response_text:
            response_text = response_text.split("```")[1]
            response_text = response_text.strip()
            if response_text.startswith("yaml"):
                response_text = response_text[4:].strip()
        try:
            return yaml.safe_load(response_text)
        except Exception as e:
            logger.error(f"Error parsing response: {response_text}: {e}")
            return None

    def find_evidence_complex(
        self, obj: Union[str, Dict], label_field=None, statement_fields: Optional[List] = None
    ) -> Dict:
        if not label_field:
            for candidates in ["name", "label", "code"]:
                if candidates in obj:
                    label_field = candidates
                    break
        if not label_field:
            raise ValueError(f"No label field found in object {obj}")
        label = obj.get(label_field, None)
        if not label:
            raise ValueError(f"No label found in object {obj}")
        new_evidences = []
        for k, v in obj.items():
            if k == label_field:
                continue
            if statement_fields and k not in statement_fields:
                continue

            def _add_evidence(input_obj: dict) -> dict:
                if "evidence" in input_obj:
                    if self.evidence_update_policy == EvidenceUpdatePolicyEnum.skip:
                        return input_obj
                q = f"Subject: {label} Value: {yaml.dump(input_obj)}"
                evidence = self.find_evidence_simple(q)
                if evidence:
                    if (
                        "evidence" in input_obj
                        and self.evidence_update_policy == EvidenceUpdatePolicyEnum.append
                    ):
                        input_obj["evidence"].append(evidence)
                    else:
                        input_obj["evidence"] = evidence
                    new_evidences.append(evidence)

            if isinstance(v, list):
                for sub_obj in v:
                    if isinstance(sub_obj, dict):
                        _add_evidence(sub_obj)
            elif isinstance(v, dict):
                _add_evidence(v)
        logger.info(f"Found {len(new_evidences)} evidence objects.")
        return obj



from typing import Dict, List


def patches_to_oak_commands(patch_dict: Dict[str, List[Dict]], ont_path: str) -> str:
    for id, patches in patch_dict.items():
        notes = ""
        changes = []
        for patch in patches:
            path = patch["path"]
            op = patch["op"]
            v = patch["value"]
            if path == "/review_notes":
                notes = patch["value"]
                if isinstance(notes, list):
                    notes = ". ".join(notes)
                notes = notes.replace('"', "'")
            elif path == "definition":
                if op == "replace":
                    changes.append(f"change definition of {id} to {v}")
        if changes:
            for change in changes:
                # print(f'runoak -i {ont_path} apply "{change}"')
                print(f'robot kgcl:apply -i {ont_path} apply -k "{change}"')
            if not notes:
                notes = f"Applying {len(changes)}"
            print(f'git commit -m "{notes}"')



"""EUtils-based wrapper for studies in NCBI."""

import logging
from dataclasses import dataclass
from typing import ClassVar, Dict, List

import yaml

from curategpt.wrappers.literature.eutils_wrapper import EUtilsWrapper

logger = logging.getLogger(__name__)


@dataclass
class NCBIBioprojectWrapper(EUtilsWrapper):
    """
    A wrapper to provide a search facade over NCBI bioproject.

    This is a dynamic wrapper: it can be used as a search facade,
    but cannot be ingested in whole.
    """

    name: ClassVar[str] = "ncbi_bioproject"

    eutils_db: ClassVar[str] = "bioproject"

    id_prefix: ClassVar[str] = "bioproject"

    default_object_type = "Study"

    def objects_from_dict(self, results: Dict) -> List[Dict]:
        objs = []
        for r in results["RecordSet"]["DocumentSummary"]:
            print(yaml.dump(r))
            p = r["Project"]
            d = p["ProjectDescr"]
            obj = {}

            obj["id"] = "bioproject:" + p["ProjectID"]["ArchiveID"]["@accession"]
            for k in ["Name", "Title"]:
                if k in d:
                    obj["title"] = d[k]
                    break
            obj["description"] = d["Description"]
            pt = p["ProjectType"]
            if "ProjectTypeSubmission" in pt:
                pts = pt["ProjectTypeSubmission"]
                if "Target" in pts:
                    if "Organism" in pts["Target"]:
                        obj["organism"] = pts["Target"]["Organism"]["OrganismName"]
            if "Publication" in d:
                pubs = d["Publication"]
                if isinstance(pubs, list):
                    obj["publications"] = [self._parse_publication(p) for p in pubs]
                else:
                    obj["publications"] = [self._parse_publication(pubs)]
            objs.append(obj)

        return objs

    def _parse_publication(self, pub: Dict) -> Dict:
        """Parse a publication from a bioproject."""
        if "StructuredCitation" in pub:
            sc = pub["StructuredCitation"]
            title = sc["Title"]
        else:
            title = None
        return {
            "title": title,
            "id": "pmid:" + pub["@id"],
        }



from dataclasses import dataclass
from typing import ClassVar

from curategpt.wrappers.clinical.hpoa_wrapper import HPOAWrapper


@dataclass
class HPOAByPubWrapper(HPOAWrapper):
    """
    A wrapper over HPOA grouping by publication
    """

    name: ClassVar[str] = "hpoa_by_pub"
    default_object_type = "Publication"
    group_by_publication: bool = True



import logging

from llm import Model, Response
from tenacity import retry, retry_if_exception, stop_after_attempt, wait_random_exponential

logger = logging.getLogger(__name__)


def is_rate_limit_error(exception):
    # List of fully qualified names of RateLimitError exceptions from various libraries
    rate_limit_errors = [
        "openai.error.RateLimitError",
        "openai.RateLimitError",
        "anthropic.error.RateLimitError",
        "anthropic.RateLimitError",
        # Add more as needed
    ]
    exception_full_name = f"{exception.__class__.__module__}.{exception.__class__.__name__}"
    logger.warning(f"Exception_full_name: {exception_full_name}")
    logger.warning(f"Exception: {exception}")
    return exception_full_name in rate_limit_errors


@retry(
    retry=retry_if_exception(is_rate_limit_error),
    wait=wait_random_exponential(multiplier=1, max=40),
    stop=stop_after_attempt(3),
)
def query_model(model: Model, *args, **kwargs) -> Response:
    logger.debug(f"Querying model {model.model_id}, args: {args}, kwargs: {kwargs}")
    response = model.prompt(*args, **kwargs)
    _text = response.text()
    return response



"""Chat with issues from a GitHub repository."""

import logging
import os
from dataclasses import dataclass
from time import sleep
from typing import ClassVar, Dict, Iterable, Iterator, List, Optional

import requests
import requests_cache
from pydantic import BaseModel, ConfigDict

from curategpt.wrappers.base_wrapper import BaseWrapper

logger = logging.getLogger(__name__)


class Comment(BaseModel):
    model_config = ConfigDict(protected_namespaces=())
    id: str
    user: str = None
    body: str = None


class PullRequest(BaseModel):
    model_config = ConfigDict(protected_namespaces=())
    id: str
    number: int = None
    title: str = None
    user: str = None
    labels: List[str] = None
    state: str = None
    assignees: List[str] = None
    created_at: str = None
    body: str = None
    comments: List[Comment] = None


class Issue(BaseModel):
    model_config = ConfigDict(protected_namespaces=())
    id: str
    number: int = None
    type: str = None
    title: str = None
    user: str = None
    labels: List[str] = None
    state: str = None
    assignees: List[str] = None
    created_at: str = None
    body: str = None
    # pull_request: str = None
    comments: List[Comment] = None


def pr_comments(self, pr_number: str) -> Iterator[Dict]:
    session = self.session
    url = f"https://api.github.com/repos/{self.owner}/{self.repo}/pulls/{pr_number}/comments"
    params = {"per_page": 100}

    while url:
        response = session.get(url, headers=self.headers, params=params)
        response.raise_for_status()
        yield from response.json()
        url = response.links.get("next", {}).get("url")


def get_token(token: str = None) -> Optional[str]:
    if token:
        return token
    """Get token from env var"""
    token = os.environ.get("CURATEGPT_GITHUB_TOKEN")
    # if not token:
    #    raise ValueError("CURATEGPT_GITHUB_TOKEN env var not set")
    return token


@dataclass
class GitHubWrapper(BaseWrapper):
    """
    A wrapper to provide a search facade over GitHub.

    This is a dynamic wrapper: it can be used as a search facade,
    but cannot be ingested in whole.
    """

    name: ClassVar[str] = "github"

    cache_name: ClassVar[str] = "github_requests"

    default_object_type = "Issue"

    session: requests.Session = None

    owner: str = None
    repo: str = None

    _repo_description: str = None

    def __post_init__(self):
        self.session = requests_cache.CachedSession(self.cache_name)
        if self.repo and "/" in self.repo:
            if self.owner:
                raise ValueError("Cannot specify both owner and a slash in repo")
            self.owner, self.repo = self.repo.split("/")

    @property
    def headers(self):
        token = get_token()
        hdr = {
            "Authorization": f"token {token}",
            "Accept": "application/vnd.github.v3+json",
            "User-Agent": "CurateGPT/0.0.1",
        }
        if not token:
            del hdr["Authorization"]
        logger.info(f"Header: {hdr}")
        return hdr

    @property
    def repo_description(self) -> str:
        if not self._repo_description:
            url = f"https://api.github.com/repos/{self.owner}/{self.repo}"
            response = requests.get(url, headers=self.headers)
            response.raise_for_status()
            repo_data = response.json()
            self._repo_description = repo_data.get("description")
        return self._repo_description

    def external_search(
            self, text: str, expand: bool = True, limit=None, token: str = None, **kwargs
    ) -> List[Dict]:
        token = get_token(token)
        if limit is None:
            limit = 10
        if expand:
            logger.info(f"Expanding search term: {text} to create query")
            model = self.extractor.model
            q = (
                "I will give you a piece of text."
                "You will generate a semi-colon separated list of the 3 most relevant terms"
                f" to search the {self.owner}/{self.repo} repo on GitHub."
                "Keep this list minimal and precise. Use semi-colons to separate terms."
                "Use terms that within the domain of the repo."
                f"The description of the repo is: {self.repo_description}.\n"
                f"\n---\nHere is the text:\n{text}"
            )
            response = model.prompt(
                q,
                system="You are an agent to expand query terms to search github. ALWAYS SEPARATE WITH SEMI-COLONS",
            )
            terms = [f'"{x.strip()}"' for x in response.text().split(";")]
            search_term = " OR ".join(terms[:3])
        else:
            search_term = text
        logger.info(f"Constructed search term: {search_term}")
        url = f"https://api.github.com/search/issues?q={search_term} repo:{self.owner}/{self.repo}&type=issues"

        headers = self.headers
        del headers["Authorization"]

        response = requests.get(url, headers=headers)
        if response.status_code != 200:
            response.raise_for_status()
        all_issues = []
        issues_dicts = response.json()["items"]
        for issue_dict in issues_dicts:
            issue_dict["comments"] = list(self.issue_comments(issue_dict["number"]))
            issue_obj = self.transform_issue(issue_dict)
            all_issues.append(issue_obj.dict())
            if len(all_issues) >= limit:
                break
        return all_issues

    def objects(
            self,
            collection: str = None,
            object_ids: Optional[Iterable[str]] = None,
            token: str = None,
            **kwargs,
    ) -> Iterator[Dict]:
        session = self.session
        token = get_token(token)
        url = f"https://api.github.com/repos/{self.owner}/{self.repo}/issues"
        headers = {"Authorization": f"token {token}", "Accept": "application/vnd.github.v3+json"}
        if not token:
            del headers["Authorization"]
            sleep(5)
        logger.debug(f"Header: {headers}")
        params = {
            "state": "all",  # To fetch both open and closed issues and PRs
            "per_page": 100,  # Fetch 100 results per page (max allowed)
        }

        while url:
            response = session.get(url, headers=headers, params=params)
            response.raise_for_status()
            issues = response.json()
            for issue in issues:
                issue_number = issue.get("number")
                # Fetch both issue comments and PR comments
                if "pull_request" in issue:
                    issue["comments"] = list(self.pr_comments(issue_number))
                else:
                    issue["comments"] = list(self.issue_comments(issue_number))
                issue_obj = self.transform_issue(issue)
                yield issue_obj.dict()
                # Check if there are more pages to process
            url = response.links.get("next", {}).get("url")
            if not response.from_cache:
                sleep(0.2)

    def issue_comments(self, issue_number: str) -> Iterator[Dict]:
        session = self.session
        url = (
            f"https://api.github.com/repos/{self.owner}/{self.repo}/issues/{issue_number}/comments"
        )
        params = {"per_page": 100}

        while url:
            response = session.get(url, headers=self.headers, params=params)
            response.raise_for_status()
            yield from response.json()
            url = response.links.get("next", {}).get("url")

    def transform_issue(self, obj: Dict) -> Issue:

        if not obj.get("body"):
            obj["body"] = ""

        issue = Issue(
            id=obj.get("url"),
            number=obj.get("number"),
            title=obj.get("title"),
            user=obj.get("user").get("login"),
            labels=[label.get("name") for label in obj.get("labels")],
            state=obj.get("state"),
            assignees=[assignee.get("login") for assignee in obj.get("assignees")],
            created_at=obj.get("created_at"),
            type="pull_request" if obj.get("pull_request") else "issue",
            body=obj.get("body"),
            # pull_request=pr.get("url") if pr else None,
            comments=[
                Comment(
                    id=comment.get("url"),
                    user=comment.get("user").get("login"),
                    body=comment.get("body"),
                )
                for comment in obj.get("comments")
            ],
        )
        return issue

    def pr_comments(self, pr_number: str) -> Iterator[Dict]:
        session = self.session
        url = f"https://api.github.com/repos/{self.owner}/{self.repo}/pulls/{pr_number}/comments"
        params = {"per_page": 100}

        while url:
            response = session.get(url, headers=self.headers, params=params)
            response.raise_for_status()
            yield from response.json()
            url = response.links.get("next", {}).get("url")



"""Chat with a KB."""

import logging
import re
from dataclasses import dataclass
from typing import Any, Dict, Optional

import yaml
from llm import Conversation
from pydantic import BaseModel, ConfigDict

from curategpt.agents.base_agent import BaseAgent
from curategpt.utils.tokens import estimate_num_tokens, max_tokens_by_model
from curategpt.wrappers import BaseWrapper

logger = logging.getLogger(__name__)


class ChatResponse(BaseModel):
    """
    Response from chat engine.

    TODO: Rename class to indicate that it is provenance-enabled chat
    """

    model_config = ConfigDict(protected_namespaces=())

    body: str
    """Text of response."""

    prompt: str
    """Prompt used to generate response."""

    formatted_body: str = None
    """Body formatted with markdown links to references."""

    references: Optional[Dict[str, Any]] = None
    """References for citations detected in response."""

    uncited_references: Optional[Dict[str, Any]] = None
    """Potential references for which there was no detected citation."""


def replace_references_with_links(text):
    """Replace references with links."""
    pattern = r"\[(\d+)\]"
    replacement = lambda m: f"[{m.group(1)}](#ref-{m.group(1)})"
    return re.sub(pattern, replacement, text)


@dataclass
class ChatAgent(BaseAgent):
    """
    An agent that allows chat to a knowledge source.

    This implements a standard knowledgebase retrieval augmented generation pattern.
    The knowledge_source is queried for relevant objects (the source can be a local
    database or a remote source such as pubmed).
    The objects are provided as context to a LLM query
    """

    relevance_factor: float = 0.5
    """Relevance factor for diversifying search results using MMR."""

    conversation_id: Optional[str] = None

    def chat(
        self,
        query: str,
        conversation: Optional[Conversation] = None,
        limit: int = 10,
        collection: str = None,
        expand=True,
        **kwargs,
    ) -> ChatResponse:
        """
        Extract structured object using text seed and background knowledge.

        :param text:
        :param kwargs:
        :return:
        """
        if self.extractor is None:
            if isinstance(self.knowledge_source, BaseWrapper):
                self.extractor = self.knowledge_source.extractor
            else:
                raise ValueError("Extractor must be set.")
        logger.info(f"Chat: {query} on {self.knowledge_source} kwargs: {kwargs}, limit: {limit}")
        if collection is None:
            collection = self.knowledge_source_collection
        kwargs["collection"] = collection
        kb_results = list(
            self.knowledge_source.search(
                query, relevance_factor=self.relevance_factor, limit=limit, expand=expand, **kwargs
            )
        )
        while True:
            i = 0
            references = {}
            texts = []
            current_length = 0
            for obj, _, _obj_meta in kb_results:
                i += 1
                obj_text = yaml.dump({k: v for k, v in obj.items() if v}, sort_keys=False)
                references[str(i)] = obj_text
                texts.append(f"## Reference {i}\n{obj_text}")
                current_length += len(obj_text)
            model = self.extractor.model
            prompt = "I will first give background facts, then ask a question. Use the background fact to answer\n"
            prompt += "---\nBackground facts:\n"
            prompt += "\n".join(texts)
            prompt += "\n\n"
            prompt += "I will ask a question and you will answer as best as possible, citing the references above.\n"
            prompt += "Write references in square brackets, e.g. [1].\n"
            prompt += (
                "For additional facts you are sure of but a reference is not found, write [?].\n"
            )
            prompt += f"---\nHere is the Question: {query}.\n"
            logger.debug(f"Candidate Prompt: {prompt}")
            estimated_length = estimate_num_tokens([prompt])
            logger.debug(
                f"Max tokens {self.extractor.model.model_id}: {max_tokens_by_model(self.extractor.model.model_id)}"
            )
            # TODO: use a more precise estimate of the length
            if estimated_length + 300 < max_tokens_by_model(self.extractor.model.model_id):
                break
            else:
                # remove least relevant
                logger.debug(f"Removing least relevant of {len(kb_results)}: {kb_results[-1]}")
                if not kb_results:
                    raise ValueError(f"Prompt too long: {prompt}.")
                kb_results.pop()

        logger.info(f"Prompt: {prompt}")

        if conversation:
            conversation.model = model
            agent = conversation
            conversation_id = conversation.id
            logger.info(f"Conversation ID: {conversation_id}")
        else:
            agent = model
            conversation_id = None
        response = agent.prompt(prompt, system="You are a scientist assistant.")
        response_text = response.text()
        pattern = r"\[(\d+|\?)\]"
        used_references = re.findall(pattern, response_text)
        used_references_dict = {ref: references.get(ref, "NO REFERENCE") for ref in used_references}
        uncited_references_dict = {
            ref: ref_obj for ref, ref_obj in references.items() if ref not in used_references
        }
        formatted_text = replace_references_with_links(response_text)
        return ChatResponse(
            body=response_text,
            formatted_body=formatted_text,
            prompt=prompt,
            references=used_references_dict,
            uncited_references=uncited_references_dict,
            conversation_id=conversation_id,
        )



import logging
from typing import Dict

import pandas as pd
import yaml
from huggingface_hub import HfApi, create_repo

logger = logging.getLogger(__name__)

class HuggingFaceAgent:
    def upload(self, objects, metadata, repo_id, private=False, **kwargs):
        """
        Upload an entire collection to a Hugging Face repository.

        :param objects: The objects to upload.
        :param metadata: The metadata associated with the collection.
        :param repo_id: The repository ID on Hugging Face.
        :param private: Whether the repository should be private.
        :param kwargs: Additional arguments such as batch size or metadata options.
        """
        venomx_metadata = self._transform_metadata_to_venomx(metadata)
        embedding_file = "embeddings.parquet"
        metadata_file = "metadata.yaml"
        pd.DataFrame(objects).to_parquet(embedding_file)
        with open(metadata_file, "w") as f:
            yaml.dump(venomx_metadata, f)

        self._create_repo(repo_id, private=private)
        self._upload_files(repo_id, {
            embedding_file: embedding_file,
            metadata_file: metadata_file,
        })

    def _create_repo(self, repo_id: str, private: bool = False):
        """
        Create a new repository on Hugging Face Hub.

        :param repo_id: The repository ID, e.g., "biomedical-translator/[your repo name]".
        :param private: Whether the repository is private.
        """
        try:
            create_repo(repo_id=repo_id, repo_type="dataset", private=private)
            logger.info(f"Repository {repo_id} created successfully on Hugging Face.")
        except Exception as e:
            logger.error(f"Failed to create repository {repo_id} on Hugging Face: {e}")
            raise

    def _upload_files(self, repo_id: str, files: Dict[str, str]):
        """
        Upload files to a Hugging Face repository.

        :param repo_id: The repository ID on Hugging Face.
        :param files: A dictionary with local file paths as keys and paths in the repository as values.
        """
        api = HfApi()
        try:
            for local_path, repo_path in files.items():
                api.upload_file(
                    path_or_fileobj=local_path,
                    path_in_repo=repo_path,
                    repo_id=repo_id,
                    repo_type="dataset",
                )
                logger.info(f"Uploaded {local_path} to {repo_path} in {repo_id}")
        except Exception as e:
            logger.error(f"Failed to upload files to {repo_id} on Hugging Face: {e}")
            raise

    def _transform_metadata_to_venomx(self, metadata):
        """
        Transform metadata from ChromaDB format to VenomX format.

        :param metadata: Metadata object from store
        :return: A dictionary formatted according to VenomX
        """

        prefixes = metadata.prefixes if hasattr(metadata,
                                                'prefixes') and metadata.prefixes else {}

        venomx_metadata = {
            "description": metadata.description or "No description provided",
            "prefixes": prefixes,
            "model": {
                "name": metadata.model or "unknown",
                # Default to a known model if not specified
            },
            "model_input_method": {
                "description": "Simple pass through of labels only",
                "fields": ["rdfs:label"]
                # Adjust fields based on actual data structure if needed
            },
            "dataset": {
                "name": metadata.name or "Unknown Dataset",
                "url": metadata.source or "Unknown URL"
                # Adjust based on available metadata
            }
        }

        # Enrich VenomX format with annotations if available
        if metadata.annotations:
            venomx_metadata["annotations"] = metadata.annotations

        # Include any additional fields from metadata that are relevant
        if hasattr(metadata, 'extra_fields') and metadata.extra_fields:
            venomx_metadata.update(metadata.extra_fields)

        return venomx_metadata



"""Simple default adapter for a object store."""

import logging
from dataclasses import dataclass, field
from typing import ClassVar, Dict, Iterable, Iterator, List, Optional, Union, get_origin

from pydantic import BaseModel, ConfigDict

from curategpt import DBAdapter
from curategpt.store.db_adapter import OBJECT, PROJECTION, QUERY, SEARCH_RESULT
from curategpt.store.metadata import CollectionMetadata

logger = logging.getLogger(__name__)


class Collection(BaseModel):

    model_config = ConfigDict(protected_namespaces=())

    objects: List[Dict] = []
    metadata: Dict = {}

    def add(self, object: Dict) -> None:
        self.objects.append(object)

    def delete(self, key_value: str, key: str) -> None:
        self.objects = [obj for obj in self.objects if obj[key] != key_value]


class CollectionIndex(BaseModel):

    model_config = ConfigDict(protected_namespaces=())

    collections: Dict[str, Collection] = {}

    def get_collection(self, name: str) -> Collection:
        if name not in self.collections:
            self.collections[name] = Collection()
        return self.collections[name]


@dataclass
class InMemoryAdapter(DBAdapter):
    """
    Simple in-memory adapter for a object store.
    """

    name: ClassVar[str] = "in_memory"

    collection_index: CollectionIndex = field(default_factory=CollectionIndex)

    # CUD operations

    def _get_collection_object(self, collection_name: str) -> Collection:
        """
        Get the collection object for a collection name.

        :param collection_name:
        :return:
        """
        collection_obj = self.collection_index.get_collection(self._get_collection(collection_name))
        return collection_obj

    def insert(self, objs: Union[OBJECT, Iterable[OBJECT]], collection: str = None, **kwargs):
        """
        Insert an object or list of objects into the store.

        :param objs:
        :param collection:
        :return:
        """
        collection_obj = self._get_collection_object(collection)
        if get_origin(type(objs)) is not Dict:
            objs = [objs]
        collection_obj.add(objs)

    def update(self, objs: Union[OBJECT, List[OBJECT]], collection: str = None, **kwargs):
        """
        Update an object or list of objects in the store.

        :param objs:
        :param collection:
        :return:
        """
        collection_obj = self._get_collection_object(collection)
        if isinstance(objs, OBJECT):
            objs = [objs]
        collection_obj.add(objs)

    def upsert(self, objs: Union[OBJECT, List[OBJECT]], collection: str = None, **kwargs):
        """
        Upsert an object or list of objects in the store.

        :param objs:
        :param collection:
        :return:
        """
        collection_obj = self._get_collection_object(collection)
        if isinstance(objs, OBJECT):
            objs = [objs]
        collection_obj.add(objs)

    def delete(self, id: str, collection: str = None, **kwargs):
        """
        Delete an object by its ID.

        :param id:
        :param collection:
        :return:
        """
        collection_obj = self._get_collection_object(collection)
        collection_obj.delete(id, self.identifier_field())

    # Collection operations

    def remove_collection(self, collection: str = None, exists_ok=False, **kwargs):
        """
        Remove a collection from the database.

        :param collection:
        :return:
        """
        if not exists_ok and collection not in self.collection_index.collections:
            raise ValueError(f"Collection {collection} does not exist.")
        self.collection_index.collections.pop(self._get_collection(collection), None)

    def list_collection_names(self) -> List[str]:
        """
        List all collections in the database.

        :return:
        """
        return list(self.collection_index.collections.keys())

    def collection_metadata(
        self, collection_name: Optional[str] = None, include_derived=False, **kwargs
    ) -> Optional[CollectionMetadata]:
        """
        Get the metadata for a collection.

        :param collection_name:
        :param include_derived: Include derived metadata, e.g. counts
        :return:
        """
        collection_obj = self._get_collection_object(collection_name)
        md_dict = collection_obj.metadata
        cm = CollectionMetadata(**md_dict)
        if include_derived:
            cm.object_count = len(collection_obj.objects)
        return cm

    def set_collection_metadata(
        self, collection_name: Optional[str], metadata: CollectionMetadata, **kwargs
    ):
        """
        Set the metadata for a collection.

        :param collection_name:
        :return:
        """
        collection_obj = self._get_collection_object(collection_name)
        collection_obj.metadata = metadata.dict()

    def update_collection_metadata(self, collection_name: str, **kwargs) -> CollectionMetadata:
        """
        Update the metadata for a collection.

        :param collection_name:
        :param kwargs:
        :return:
        """
        raise NotImplementedError

    # Query operations

    def search(
        self, text: str, where: QUERY = None, collection: str = None, **kwargs
    ) -> Iterator[SEARCH_RESULT]:
        """
        Query the database for a text string.

        :param text:
        :param collection:
        :param where:
        :param kwargs:
        :return:
        """
        raise NotImplementedError

    def find(
        self,
        where: QUERY = None,
        projection: PROJECTION = None,
        collection: str = None,
        **kwargs,
    ) -> Iterator[SEARCH_RESULT]:
        """
        Query the database.

        :param text:
        :param collection:
        :param kwargs:
        :return:
        """
        raise NotImplementedError

    def matches(self, obj: OBJECT, **kwargs) -> Iterator[SEARCH_RESULT]:
        """
        Query the database for matches to an object.

        :param obj:
        :param kwargs:
        :return:
        """
        raise NotImplementedError

    def lookup(self, id: str, collection: str = None, **kwargs) -> OBJECT:
        """
        Lookup an object by its ID.

        :param id:
        :param collection:
        :return:
        """
        collection_obj = self._get_collection_object(collection)
        return collection_obj.get(id, self.identifier_field())

    def peek(self, collection: str = None, limit=5, **kwargs) -> Iterator[OBJECT]:
        """
        Peek at first N objects in a collection.

        :param collection:
        :param limit:
        :return:
        """
        collection_obj = self._get_collection_object(collection)
        yield from collection_obj.objects[:limit]

    def fetch_all_objects_memory_safe(
        self, collection: str = None, batch_size: int = 100, **kwargs
    ) -> Iterator[OBJECT]:
        pass



"""Chat with a KB."""

import logging
import time
from dataclasses import dataclass, field
from typing import ClassVar, Dict, List, Optional

import requests
import requests_cache
from oaklib import BasicOntologyInterface

from curategpt.wrappers import BaseWrapper

logger = logging.getLogger(__name__)

BASE_URL = "https://clinicaltrials.gov/api/v2"
STUDY_URL = f"{BASE_URL}/studies"

RATE_LIMIT_DELAY = 0.5


@dataclass
class ClinicalTrialsWrapper(BaseWrapper):
    """
    A wrapper over a clinicaltrials.gov.

    """

    name: ClassVar[str] = "ctgov"

    _label_adapter: BasicOntologyInterface = None

    default_object_type = "ClinicalTrial"

    session: requests.Session = field(default_factory=lambda: requests.Session())

    where: Optional[Dict] = None

    _uses_cache: bool = False

    def set_cache(self, name: str) -> None:
        self.session = requests_cache.CachedSession(name)
        self._uses_cache = True

    def external_search(
        self, text: str, expand: bool = True, where: Optional[Dict] = None, **kwargs
    ) -> List:
        if expand:
            logger.info(f"Expanding search term: {text} to create ctgov query")
            model = self.extractor.model
            response = model.prompt(
                text,
                system=(
                    "generate a MINIMAL semi-colon separated list of the most relevant terms. "
                    "make terms general such that when they are concatenated with AND they can be "
                    "used to search for the desired clinical trial."
                ),
            )
            terms = response.text().split(";")
            terms = [f'"{term}"' for term in terms]
            search_term = " AND ".join(terms)
        else:
            search_term = text
        if not where and self.where:
            where = self.where
        if where:
            whereq = " AND ".join([f"{v}[{k}]" for k, v in where.items()])
            search_term = f"({search_term}) AND {whereq}"
        logger.info(f"Constructed search term: {search_term}")
        # Parameters for the request
        params = {
            "query.term": search_term,
        }
        # Note: we don't cache this call as there could be many
        # different search terms
        response = requests.get(STUDY_URL, params=params)
        time.sleep(RATE_LIMIT_DELAY)
        data = response.json()
        return self.objects_from_list(data["studies"])

    def objects_from_list(self, input_objs: List[Dict]) -> List[Dict]:
        objs = []
        for input_obj in input_objs:
            protocolSection = input_obj["protocolSection"]
            identificationModule = protocolSection["identificationModule"]
            # print(protocolSection)
            obj = {
                "id": identificationModule["nctId"],
                "title": identificationModule["briefTitle"],
                "description": protocolSection["descriptionModule"],
            }
            objs.append(obj)
        return objs



"""Chat with a KB."""

import logging
from dataclasses import dataclass
from typing import ClassVar, Dict, List

from curategpt.wrappers.literature.eutils_wrapper import EUtilsWrapper

logger = logging.getLogger(__name__)


@dataclass
class NCBIBiosampleWrapper(EUtilsWrapper):
    """
    A wrapper to provide a search facade over NCBI Biosample.

    This is a dynamic wrapper: it can be used as a search facade,
    but cannot be ingested in whole.
    """

    name: ClassVar[str] = "ncbi_biosample"

    eutils_db: ClassVar[str] = "biosample"

    id_prefix: ClassVar[str] = "biosample"

    default_object_type = "Sample"

    def objects_from_dict(self, results: Dict) -> List[Dict]:
        samples = []
        for s in results["BioSampleSet"]["BioSample"]:
            d = s["Description"]
            sample = {}
            sample["id"] = f"biosample:{s['@accession']}"
            sample["title"] = d["Title"]
            sample["organism"] = d["Organism"]["@taxonomy_name"]
            sample["package"] = s["Package"]["@display_name"]
            if not s["Attributes"]:
                logger.warning(f"Skipping sample with no attributes: {s}")
                continue
            for a in s["Attributes"]["Attribute"]:
                if isinstance(a, str):
                    logger.warning(f"Skipping attribute: {a} in {s['Attributes']}")
                    continue
                a_name = a.get("@harmonized_name", a.get("@attribute_name"))
                sample[a_name] = a["#text"]
            samples.append(sample)

        return samples



from enum import Enum
from typing import List

import streamlit as st
from matplotlib import pyplot as plt
from sklearn.decomposition import PCA
from sklearn.manifold import TSNE

DEFAULT_LIMIT_SLIDER_HELP = """
Examples that are similar to your query are picked from the selected
knowledge base, and used as context to guide the LLM.
If you pick too many examples, it may go beyond the limits of the context window
for the model you selected.
"""


class DimensionalityReductionOptions(str, Enum):
    PCA = "PCA"
    TSNE = "t-SNE"
    UMAP = "UMAP"


def limit_slider_component(name="Max examples", tooltip=DEFAULT_LIMIT_SLIDER_HELP):
    return st.slider(
        name,
        min_value=0,
        max_value=20,
        value=10,
        step=1,
        help=tooltip,
    )


def vectors_to_fig(labels: List[str], vectors: List, method: DimensionalityReductionOptions = None):
    if method == DimensionalityReductionOptions.PCA:
        reducer = PCA(n_components=2)
    elif method is None or method == DimensionalityReductionOptions.TSNE:
        n_samples = len(vectors)
        perplexity_value = min(
            n_samples - 1, 30
        )  # Default is 30, but should be less than number of samples
        reducer = TSNE(n_components=2, perplexity=perplexity_value)
    elif method == DimensionalityReductionOptions.UMAP:
        # TODO: umap-learn is hard to install on a mac
        raise NotImplementedError("UMAP not yet implemented")
        # reducer = umap.UMAP()
    else:
        raise ValueError(f"Unknown method {method}")

    reduced_data = reducer.fit_transform(vectors)

    fig, ax = plt.subplots(figsize=(12, 10))
    ax.scatter(reduced_data[:, 0], reduced_data[:, 1], s=50)
    for i, label in enumerate(labels):
        ax.annotate(label, (reduced_data[i, 0], reduced_data[i, 1]), fontsize=9, ha="right")
    return fig



import logging
from dataclasses import dataclass
from typing import ClassVar, Dict, Iterable, Iterator, Optional

import requests_cache
from oaklib import BasicOntologyInterface

from curategpt.wrappers import BaseWrapper

logger = logging.getLogger(__name__)

BASE_URL = "https://mediadive.dsmz.de/rest"


@dataclass
class MediaDiveWrapper(BaseWrapper):
    """
    A wrapper over MediaDive.
    """

    name: ClassVar[str] = "mediadive"

    _label_adapter: BasicOntologyInterface = None

    default_object_type = "ComplexMixture"

    def object_ids(self, **kwargs) -> Iterator[str]:
        """
        Get all medium ids in database

        :param kwargs:
        :return:
        """
        session = requests_cache.CachedSession("mediadive")

        response = session.get(f"{BASE_URL}/media")
        response.raise_for_status()
        data = response.json()["data"]
        yield from [obj["id"] for obj in data]

    def objects(
        self, collection: str = None, object_ids: Optional[Iterable[str]] = None, **kwargs
    ) -> Iterator[Dict]:
        """
        All media

        :param collection:
        :param object_ids: optional list of object ids to retrieve (if None, all objects are retrieved)
        :param kwargs:
        :return:
        """
        session = requests_cache.CachedSession("mediadive")

        if not object_ids:
            object_ids = self.object_ids(**kwargs)
        else:
            object_ids = object_ids

        for object_id in object_ids:
            response = session.get(f"{BASE_URL}/medium/{object_id}")
            response.raise_for_status()
            data = response.json()["data"]
            obj = data["medium"]
            # TODO: Finalize after https://github.com/biopragmatics/bioregistry/issues/941
            obj["id"] = f"mediadive.medium:{obj['id']}"
            solutions = data.get("solutions", {})
            if solutions:
                obj["solutions"] = solutions
            else:
                logger.warning(f"No solutions for {object_id}")
            yield obj



from typing import Dict

import inflection
import yaml


def camelify(text: str) -> str:
    """
    Convert text to camel case.

    :param text:
    :return:
    """
    # replace all non-alphanumeric characters with underscores
    safe = "".join([c if c.isalnum() else "_" for c in text])
    return inflection.camelize(safe)


def object_as_yaml(obj: Dict) -> str:
    """
    Canonical YAML representation of an object.

    - no empty values
    - no sorting (default insertion order)

    :param obj:
    :return:
    """
    return yaml.dump({k: v for k, v in obj.items() if v}, sort_keys=False)


def remove_formatting(text: str, expect_format: str = "") -> str:
    """
    Remove markdown formatting from text if present.

    :param text:
    :param expect_format: The expected format of the text, e.g., "json" (optional)
    :return:
    """
    if text.startswith("```" + expect_format):
        return text[3 + len(expect_format) : -3]
    return text



"""Agent Utilities."""

import logging
from typing import Dict, List, Optional, Tuple

import yaml
from llm import Model

from curategpt.store.db_adapter import SEARCH_RESULT
from curategpt.utils.tokens import estimate_num_tokens, max_tokens_by_model

logger = logging.getLogger(__name__)


def select_from_options_prompt(
    kb_results: List[SEARCH_RESULT],
    model: Model,
    obj_type: str = "Reference",
    query: Optional[str] = None,
    prompt_template: Optional[str] = None,
    id_field: str = None,
) -> Tuple[str, Dict[str, str], Dict]:
    """
    Prompt user to select from a list of options.

    :param kb_results: order from most relevant
    :param model:
    :param obj_type:
    :param query:
    :param prompt_template:
    :return:
    """
    if prompt_template is None:
        if query is None:
            raise ValueError("Either query or prompt_template must be specified.")
        prompt_template = "I will first give background facts, then ask a question."
        prompt_template += "Use the background fact to answer\n"
        prompt_template += "---\nBackground facts:\n"
        prompt_template += "\n {join(texts)}"
        prompt_template += "\n\n"
        prompt_template += "I will ask a question and you will answer as best as possible,"
        prompt_template += "citing the references above.\n"
        prompt_template += "Write references in square brackets, e.g. [1].\n"
        prompt_template += (
            "For additional facts you are sure of but a reference is not found, write [?].\n"
        )
        prompt_template += "---\nHere is the Question: {query}.\n"
    while True:
        i = 0
        references = {}
        objects = {}
        texts = []
        current_length = 0
        for obj, _, _obj_meta in kb_results:
            i += 1
            obj_text = yaml.dump({k: v for k, v in obj.items() if v}, sort_keys=False)
            references[str(i)] = obj_text
            objects[str(i)] = obj
            if id_field and id_field in obj:
                ref = obj[id_field]
            else:
                ref = f"{obj_type} {i}"
            texts.append(f"## {ref}\n{obj_text}")
            current_length += len(obj_text)
        prompt = prompt_template.format(body="".join(texts), query=query)
        logger.info(f"Prompt: {prompt}")
        estimated_length = estimate_num_tokens([prompt])
        logger.debug(f"Max tokens {model.model_id}: {max_tokens_by_model(model.model_id)}")
        # TODO: use a more precise estimate of the length
        if estimated_length + 300 < max_tokens_by_model(model.model_id):
            break
        else:
            # remove least relevant
            if not kb_results:
                raise ValueError(f"Prompt too long: {prompt}.")
            kb_results.pop()
    return prompt, references, objects



"""Wrapper for JSON (or YAML) documents."""

import json
import logging
from dataclasses import dataclass
from typing import ClassVar, Dict, Iterable, Iterator, Optional

from curategpt.wrappers.base_wrapper import BaseWrapper

logger = logging.getLogger(__name__)


@dataclass
class BacDiveWrapper(BaseWrapper):
    """
    A wrapper over a bacdive json files.

    This is a static wrapper: it cannot be searched
    """

    name: ClassVar[str] = "bacdive"
    prefix = "bacdive"
    default_object_type = "OrganismTaxon"
    format: str = None
    path_expression: str = None

    def objects(
        self, collection: str = None, object_ids: Optional[Iterable[str]] = None, **kwargs
    ) -> Iterator[Dict]:
        path = self.source_locator

        with open(path) as file:
            data = json.load(file)
            for obj in data.values():
                yield from self.wrap_object(obj)

    def wrap_object(self, obj: Dict) -> Iterator[Dict]:
        general = obj["General"]
        name_info = obj["Name and taxonomic classification"]
        new_obj = {}
        new_obj["id"] = self.create_curie(general["BacDive-ID"])
        new_obj["name"] = name_info.get("full scientific name", None)
        if not new_obj["name"]:
            new_obj["name"] = name_info["LPSN"].get("scientific name", None)
        taxs = general.get("NCBI tax id", None)
        if not isinstance(taxs, list):
            taxs = [taxs]
        while taxs:
            tax = taxs.pop()
            if tax:
                tax_id = tax.get("NCBI tax id", None)
                new_obj["taxon"] = f"NCBITaxon:{tax_id}"
                break
        new_obj = {**new_obj, **obj}
        yield new_obj



from dataclasses import dataclass
from pathlib import Path
from typing import Dict, Optional, Union

from linkml_runtime import SchemaView
from linkml_runtime.linkml_model import SchemaDefinition
from pydantic import BaseModel, ConfigDict


@dataclass
class SchemaProxy:
    """
    Manage connection to a schema
    """

    model_config = ConfigDict(protected_namespaces=())

    schema_source: Union[str, Path, SchemaDefinition] = None
    _pydantic_root_model: BaseModel = None
    _schemaview: SchemaView = None

    @property
    def pydantic_root_model(self) -> BaseModel:
        """
        Get the pydantic root model.

        If none is set, then generate it from the schema.

        :return:
        """
        if self._pydantic_root_model is None:
            from linkml.generators.pydanticgen import PydanticGenerator

            mod = PydanticGenerator(self.schema).compile_module()
            roots = [c.name for c in self.schemaview.all_classes().values() if c.tree_root]
            if not roots:
                roots = list(self.schemaview.all_classes(imports=False).keys())
            if not roots:
                raise ValueError(f"Cannot find root class in {self.schema.name}")
            if len(roots) > 1:
                raise ValueError(f"Multiple roots found in {self.schema.name}: {roots}")
            root = roots[0]
            self._pydantic_root_model = mod.__dict__[root]
        return self._pydantic_root_model

    @pydantic_root_model.setter
    def pydantic_root_model(self, value: BaseModel):
        self._pydantic_root_model = value

    @property
    def schemaview(self) -> SchemaView:
        """
        Get the schema view.

        :return:
        """
        if self._schemaview is None:
            self._schemaview = SchemaView(self.schema_source)
        return self._schemaview

    @property
    def schema(self) -> SchemaDefinition:
        """
        Get the schema

        :return:
        """
        return self.schemaview.schema

    def json_schema(self) -> Dict:
        """
        Get the JSON schema translation of the schema.
        :return:
        """
        return self.pydantic_root_model.schema()

    @property
    def name(self) -> Optional[str]:
        """
        Get the name of the schema.

        :return:
        """
        if self.schema:
            return self.schema.name
        if self.pydantic_root_model:
            return self.pydantic_root_model.__name__



"""Chat with a Google Drive."""

import logging
import os
from dataclasses import dataclass
from typing import Any, ClassVar, Dict, List

from google.auth.transport.requests import Request
from google.oauth2.credentials import Credentials
from google_auth_oauthlib.flow import InstalledAppFlow
from googleapiclient.discovery import build

from curategpt.wrappers.base_wrapper import BaseWrapper

logger = logging.getLogger(__name__)


# Authenticate and build the service
SCOPES = ["https://www.googleapis.com/auth/drive.readonly"]


@dataclass
class GoogleDriveWrapper(BaseWrapper):
    """
    A wrapper to provide a search facade over Wikipedia.

    This is a dynamic wrapper: it can be used as a search facade,
    but cannot be ingested in whole.
    """

    name: ClassVar[str] = "google_drive"

    google_drive_id: str = "0ABiscjZCrxjUUk9PVA"
    google_folder_id: str = None

    default_doc_types = ["application/vnd.google-apps.document"]

    service: Any = None

    search_limit_multiplier: ClassVar[int] = 1

    def __post_init__(self):
        creds = None
        if os.path.exists("token.json"):
            creds = Credentials.from_authorized_user_file("token.json")

        if not creds or not creds.valid:
            if creds and creds.expired and creds.refresh_token:
                creds.refresh(Request())
            else:
                flow = InstalledAppFlow.from_client_secrets_file("credentials.json", SCOPES)
                creds = flow.run_local_server(port=0)
                with open("token.json", "w") as token:
                    token.write(creds.to_json())
        self.service = build("drive", "v3", credentials=creds)

    def external_search(self, text: str, expand: bool = False, limit=None, **kwargs) -> List[Dict]:
        if limit is None:
            limit = 10
        if expand:
            logger.warning("Google Drive does not support expansion")
        service = self.service
        query = f"fullText contains '{text}' and mimeType = 'application/vnd.google-apps.document'"
        response = (
            service.files()
            .list(
                q=query,
                spaces="drive",
                corpora="drive",
                driveId=self.google_drive_id,
                fields="nextPageToken, files(id, name, mimeType)",
                includeItemsFromAllDrives=True,  # Add this line
                supportsAllDrives=True,
                # Add this line too
                pageToken=None,
            )
            .execute()
        )
        logger.debug(response)
        files = list(response.get("files", []))
        files = [file for file in files if file.get("mimeType", None) in self.default_doc_types]
        logger.info(f"Search returned {len(files)} files")
        if len(files) >= limit:
            files = files[:limit]  # Trim the results if necessary
            logger.info(f"Truncating to top {limit} results")
        return self.objects_from_files(files)

    def objects_from_files(self, files: List[Dict]) -> List[Dict]:
        service = self.service
        objs = []
        for file in files:
            file_id = file["id"]
            request = service.files().export_media(fileId=file_id, mimeType="text/plain")
            response = request.execute()
            response_text = response.decode("utf-8")
            objs.append(
                {
                    "id": file_id,
                    "url": f"https://docs.google.com/document/d/{file_id}/",
                    "label": file.get("name", ""),
                    "text": response_text,
                    "mime_type": file.get("mimeType", None),
                }
            )
        return self.split_objects(objs)

    def objects_by_ids(self, object_ids: List[str]) -> List[Dict]:
        files = [{"id: object_id"} for object_id in object_ids]
        return self.objects_from_files(files)



"""Chat with a KB."""

import gzip
import logging
from dataclasses import dataclass
from typing import ClassVar, Dict, Iterable, Iterator, Optional

from oaklib import BasicOntologyInterface, get_adapter

from curategpt.formatters.format_utils import camelify
from curategpt.wrappers import BaseWrapper

logger = logging.getLogger(__name__)


RMAP = {"CID": "induces"}


@dataclass
class BiocWrapper(BaseWrapper):
    """
    A wrapper over a bioc source file.

    This provides an exhaustive list of objects.
    """

    name: ClassVar[str] = "bioc"

    _label_adapter: BasicOntologyInterface = None

    default_object_type = "AnnotatedPublication"

    def objects(
        self, collection: str = None, object_ids: Optional[Iterable[str]] = None, **kwargs
    ) -> Iterator[Dict]:
        from bioc import biocxml

        with gzip.open(str(self.source_locator), "rb") as f:
            collection = biocxml.load(f)
            for document in collection.documents:
                doc = {}
                for p in document.passages:
                    # title and abstract are the only two passages
                    doc[p.infons["type"]] = p.text
                rels = []
                for r in document.relations:
                    obj = {}
                    for k, v in r.infons.items():
                        if k == "relation":
                            obj["relation"] = RMAP.get(v, v)
                        else:
                            mesh_id = f"MESH:{v}"
                            label = self.label_adapter.label(mesh_id)
                            if not label:
                                logger.warning(f"Could not find label for {mesh_id}")
                            obj[k] = camelify(label)
                    rels.append(obj)
                doc["statements"] = rels
                yield doc

    @property
    def label_adapter(self) -> BasicOntologyInterface:
        """Get the label adapter."""
        if self._label_adapter is None:
            self._label_adapter = get_adapter("sqlite:obo:mesh")
        return self._label_adapter



"""ChromaDB adapter."""

import json
import logging
import os
import time
from dataclasses import dataclass, field
from typing import Callable, ClassVar, Iterable, Iterator, List, Mapping, Optional, Union

import chromadb
import yaml
from chromadb import ClientAPI as API
from chromadb import Settings
from chromadb.api import EmbeddingFunction
from chromadb.types import Collection
from chromadb.utils import embedding_functions
from linkml_runtime.dumpers import json_dumper
from linkml_runtime.utils.yamlutils import YAMLRoot
from oaklib.utilities.iterator_utils import chunk
from pydantic import BaseModel

from curategpt.store.db_adapter import DBAdapter
from curategpt.store.metadata import CollectionMetadata
from curategpt.store.vocab import OBJECT, PROJECTION, QUERY, SEARCH_RESULT
from curategpt.utils.vector_algorithms import mmr_diversified_search

logger = logging.getLogger(__name__)


@dataclass
class ChromaDBAdapter(DBAdapter):
    """
    An Adapter that wraps a ChromaDB client
    """

    name: ClassVar[str] = "chromadb"
    default_model = "all-MiniLM-L6-v2"
    client: API = None
    id_field: str = field(default="id")
    text_lookup: Optional[Union[str, Callable]] = field(default="text")
    id_to_object: Mapping[str, OBJECT] = field(default_factory=dict)

    default_max_document_length: ClassVar[int] = 6000  # TODO: use tiktoken

    def __post_init__(self):
        if not self.path:
            self.path = "./db"
        logger.info(f"Using ChromaDB at {self.path}")
        self.client = chromadb.PersistentClient(
            path=str(self.path), settings=Settings(allow_reset=True, anonymized_telemetry=False)
        )
        logger.info(f"ChromaDB client: {self.client}")

    def _get_collection_object(self, collection: str = None):
        return self.client.get_collection(name=self._get_collection(collection))

    def _text(self, obj: OBJECT, text_field: Union[str, Callable]):
        if isinstance(obj, list):
            raise ValueError(f"Cannot handle list of text fields: {obj}")
        if text_field is None or (isinstance(text_field, str) and text_field not in obj):
            obj = {k: v for k, v in obj.items() if v}
            t = yaml.safe_dump(obj, sort_keys=False)
        elif isinstance(text_field, Callable):
            t = text_field(obj)
        elif isinstance(obj, dict):
            t = obj[text_field]
        else:
            t = getattr(obj, text_field)
        t = t.strip()
        if not t:
            raise ValueError(f"Text field {text_field} is empty for {type(obj)} : {obj}")
        if len(t) > self.default_max_document_length:
            logger.warning(f"Truncating text field {text_field} for {str(obj)[0:100]}...")
            t = t[: self.default_max_document_length]
        return t

    def _id(self, obj: OBJECT, id_field: str):
        if isinstance(obj, dict):
            id = obj.get(id_field, None)
        else:
            id = getattr(obj, id_field, None)
        if not id:
            id = str(obj)
        self.id_to_object[id] = obj
        return id

    def _dict(self, obj: OBJECT):
        if isinstance(obj, dict):
            return obj
        elif isinstance(obj, BaseModel):
            return obj.dict(exclude_unset=True)
        elif isinstance(obj, YAMLRoot):
            return json_dumper.to_dict(obj)
        else:
            raise ValueError(f"Cannot convert {obj} to dict")

    def _object_metadata(self, obj: OBJECT):
        """
        Transform an object into metadata suitable for storage in chromadb.

        chromadb does not allow nested objects, so in addition to storing the
        top level keys that are primitive, we also store a json representation
        of the entire object at the top level with the key "_json".

        :param obj:
        :return:
        """
        dict_obj = self._dict(obj)
        dict_obj["_json"] = json.dumps(dict_obj)
        return {
            k: v for k, v in dict_obj.items() if not isinstance(v, (dict, list)) and v is not None
        }

    def reset(self):
        """
        Reset/delete the database.
        """
        self.client.reset()

    def _embedding_function(self, model: str = None) -> EmbeddingFunction:
        """
        Get the embedding function for a given model.

        :param model:
        :return:
        """
        if model is None:
            raise ValueError("Model must be specified")
        if model.startswith("openai:"):
            return embedding_functions.OpenAIEmbeddingFunction(
                api_key=os.environ.get("OPENAI_API_KEY"),
                model_name="text-embedding-ada-002",
            )
        return embedding_functions.SentenceTransformerEmbeddingFunction(model_name=model)

    def insert(
        self,
        objs: Union[OBJECT, Iterable[OBJECT]],
        **kwargs,
    ):
        self._insert_or_update(objs, method_name="add", **kwargs)

    def _insert_or_update(
        self,
        objs: Union[OBJECT, Iterable[OBJECT]],
        collection: str = None,
        batch_size: int = None,
        method_name="add",
        object_type: str = None,
        model: str = None,
        text_field: Union[str, Callable] = None,
        **kwargs,
    ):
        """
        Insert an object or list of objects into the database.

        :param objs:
        :param collection:
        :param kwargs:
        :return:
        """
        client = self.client
        collection = self._get_collection(collection)
        cm = self.collection_metadata(collection)
        if model is None:
            if cm:
                model = cm.model
            if model is None:
                model = self.default_model
        cm = self.update_collection_metadata(collection, model=model, object_type=object_type)
        ef = self._embedding_function(cm.model)
        # cm = CollectionMetadata(name=collection, model=self.model, object_type=object_type)
        cm_dict = cm.dict(exclude_none=True)
        collection_obj = client.get_or_create_collection(
            name=collection,
            embedding_function=ef,
            metadata=cm_dict,
        )
        if self._is_openai(collection_obj) and batch_size is None:
            # TODO: see https://github.com/chroma-core/chroma/issues/709
            batch_size = 100
        if batch_size is None:
            batch_size = 100000
        if text_field is None:
            text_field = self.text_lookup
        id_field = self.identifier_field(collection)
        # see https://github.com/chroma-core/chroma/issues/709
        num_objs = len(objs) if isinstance(objs, list) else "?"
        cumulative_len = 0
        for next_objs in chunk(objs, batch_size):
            next_objs = list(next_objs)
            logger.info("Preparing batch from position ...")
            docs = [self._text(o, text_field) for o in next_objs]
            docs_len = sum([len(d) for d in docs])
            cumulative_len += docs_len
            # TODO: use tiktoken to get a better estimate
            if self._is_openai(collection_obj) and cumulative_len > 3000000:
                logger.warning(f"Cumulative length = {cumulative_len}, pausing ...")
                # TODO: this is too conservative; it should be based on time of start of batch
                time.sleep(60)
                cumulative_len = 0
            logger.debug(f"Example doc (tf={text_field}): {docs[0]}")
            logger.info("Preparing metadatas...")
            metadatas = [self._object_metadata(o) for o in next_objs]
            logger.info("Preparing ids...")
            ids = [self._id(o, id_field) for o in next_objs]
            logger.info(f"Inserting {len(next_objs)} / {num_objs} objects into {collection}")
            method = getattr(collection_obj, method_name)
            method(
                documents=docs,
                metadatas=metadatas,
                ids=ids,
            )

    def update(self, objs: Union[OBJECT, List[OBJECT]], **kwargs):
        """
        Update an object or list of objects in the store.

        :param objs:
        :param collection:
        :return:
        """
        self._insert_or_update(objs, method_name="update", **kwargs)

    def upsert(self, objs: Union[OBJECT, List[OBJECT]], **kwargs):
        """
        Update an object or list of objects in the store.

        :param objs:
        :param collection:
        :return:
        """
        self._insert_or_update(objs, method_name="upsert", **kwargs)

    def remove_collection(self, collection: str = None, exists_ok=False, **kwargs):
        """
        Remove a collection from the database.

        :param collection:
        :param exists_ok:
        :return:
        """
        if not any(cn for cn in self.list_collection_names() if cn == collection):
            if not exists_ok:
                raise ValueError(f"Collection {collection} does not exist")
            return
        self.client.delete_collection(name=collection)

    def _unjson(self, obj: Mapping):
        if not obj:
            raise ValueError(f"Cannot convert {obj} to dict")
        return json.loads(obj["_json"])

    def list_collection_names(self) -> List[str]:
        """
        List all collections in the database.

        :return:
        """
        return [c.name for c in self.client.list_collections()]

    def collection_metadata(
        self, collection_name: Optional[str] = None, include_derived=False, **kwargs
    ) -> Optional[CollectionMetadata]:
        """
        Get the metadata for a collection.

        :param collection_name:
        :return:

        Parameters
        ----------
        """
        collection_name = self._get_collection(collection_name)
        try:
            logger.info(f"Getting collection object {collection_name}")
            collection_obj = self.client.get_collection(name=collection_name)
        except Exception:
            return None
        cm = CollectionMetadata(**collection_obj.metadata)
        if include_derived:
            logger.info(f"Getting object count for {collection_name}")
            cm.object_count = collection_obj.count()
        return cm

    def set_collection_metadata(
        self, collection_name: Optional[str], metadata: CollectionMetadata, **kwargs
    ):
        """
        Set the metadata for a collection.

        :param collection_name:
        :param metadata:
        :return:
        """
        self.update_collection_metadata(
            collection_name=collection_name, **metadata.dict(exclude_none=True)
        )

    def update_collection_metadata(self, collection_name: str, **kwargs) -> CollectionMetadata:
        """
        Update the metadata for a collection.

        :param collection_name:
        :param kwargs:
        :return:
        """
        collection_name = self._get_collection(collection_name)
        metadata = self.collection_metadata(collection_name=collection_name)
        if metadata is None:
            metadata = CollectionMetadata(**kwargs)
        else:
            prev_model = metadata.model
            metadata = metadata.copy(update=kwargs)
            if prev_model and metadata.model != prev_model:
                if self.client.get_or_create_collection(name=collection_name).count() > 0:
                    raise ValueError(f"Cannot change model from {prev_model} to {metadata.model}")
                else:
                    logger.info(
                        f"Changing (empty collection) model from {prev_model} to {metadata.model}"
                    )
        # self.set_collection_metadata(collection_name=collection_name, metadata=metadata)
        if metadata.name:
            assert metadata.name == collection_name
        else:
            metadata.name = collection_name
        metadata.hnsw_space = "cosine"
        self.client.get_or_create_collection(
            name=collection_name, metadata=metadata.dict(exclude_none=True)
        )
        return metadata

    def search(self, text: str, **kwargs) -> Iterator[SEARCH_RESULT]:
        yield from self._search(text=text, **kwargs)

    def _search(
        self,
        text: str = None,
        where: QUERY = None,
        collection: str = None,
        limit=10,
        include=None,
        relevance_factor: float = None,
        expand: bool = None,
        **kwargs,
    ) -> Iterator[SEARCH_RESULT]:
        logger.info(f"Searching for {text} in {collection}")
        if relevance_factor is not None and relevance_factor < 1.0:
            yield from self.diversified_search(
                text=text,
                where=where,
                collection=collection,
                limit=limit,
                relevance_factor=relevance_factor,
                **kwargs,
            )
            return
        if not include:
            include = ["metadatas", "documents", "distances"]
        if "*" in include:
            include = ["metadatas", "documents", "distances", "embeddings"]
        client = self.client
        # Note: with chromadb it is necessary to get the collection again;
        # the first time we do not know the embedding function, but do not
        # want to accidentally set it
        collection = client.get_collection(name=self._get_collection(collection))
        metadata = collection.metadata
        collection = client.get_collection(
            name=collection.name, embedding_function=self._embedding_function(metadata["model"])
        )
        logger.debug(f"Collection metadata: {metadata}")
        if text:
            query_texts = [text]
        else:
            # TODO: use get()
            query_texts = ["any"]
        if limit is not None:
            kwargs["n_results"] = limit
        logger.debug(
            f"Query texts: {query_texts} where: {where} include: {include}, kwargs={kwargs}"
        )
        if query_texts == ["any"] and "n_results" not in kwargs and False:
            results = collection.get(where=where, include=include, **kwargs)
        else:
            results = collection.query(
                query_texts=query_texts, where=where, include=include, **kwargs
            )
        metadatas = results["metadatas"][0]
        distances = results["distances"][0]
        documents = results["documents"][0]
        if "embeddings" in include:
            embeddings = results["embeddings"][0]
        else:
            embeddings = None
        for i in range(0, len(documents)):
            if embeddings:
                embeddings_i = embeddings[i]
            else:
                embeddings_i = None
            if not metadatas[i]:
                logger.error(
                    f"Empty metadata for item {i} [num: {len(metadatas)}] doc: {documents[i]}"
                )
                continue
            yield self._unjson(metadatas[i]), distances[i], {
                "embeddings": embeddings_i,
                "document": documents[i],
            }

    def find(
        self,
        where: QUERY = None,
        projection: PROJECTION = None,
        collection: str = None,
        **kwargs,
    ) -> Iterator[SEARCH_RESULT]:
        # TODO: use get
        # yield from self.search("", where=where, collection=collection, **kwargs)
        # return
        client = self.client
        collection_obj = client.get_collection(name=self._get_collection(collection))
        logger.debug(f"Finding: {collection} W={where} kwargs={kwargs}")
        results = collection_obj.get(where=where, **kwargs)
        logger.debug("Found items")
        metadatas = results["metadatas"]
        documents = results["documents"]
        if "embeddings" in results:
            embeddings = results["embeddings"]
        else:
            embeddings = None
        logger.debug(f"Found {len(documents)} items")
        for i in range(0, len(documents)):
            if not metadatas[i]:
                logger.error(
                    f"Empty metadata for item {i} [num: {len(metadatas)}] doc: {documents[i]}"
                )
                continue
            obj = (
                self._unjson(metadatas[i]),
                0.0,
                {
                    "document": documents[i],
                },
            )
            if embeddings:
                obj[2]["_embeddings"] = embeddings[i]
            yield obj

    def diversified_search(
        self,
        text: str = None,
        limit: int = None,
        relevance_factor=0.5,
        collection: str = None,
        **kwargs,
    ) -> Iterator[SEARCH_RESULT]:
        if limit is None:
            limit = 10
        logger.debug(
            f"diversified search RF={relevance_factor}, text={text}, limit={limit}, collection={collection}"
        )
        collection_obj = self._get_collection_object(collection)
        metadata = collection_obj.metadata
        ef = self._embedding_function(metadata["model"])
        if len(text) > self.default_max_document_length:
            logger.warning(
                f"Text too long ({len(text)}), truncating to {self.default_max_document_length}"
            )
            text = text[: self.default_max_document_length]

        query_embedding = ef([text])
        kwargs["include"] = ["metadatas", "documents", "distances", "embeddings"]
        logger.debug(
            f"Diversified search for '{text}' in {collection}, limit={limit}, kwargs={kwargs}"
        )
        ranked_results = list(self.search(text, limit=limit * 10, collection=collection, **kwargs))
        if not ranked_results:
            return
        import numpy as np

        rows = [np.array(r[2]["embeddings"]) for r in ranked_results]
        query = np.array(query_embedding[0])
        reranked_indices = mmr_diversified_search(
            query, rows, relevance_factor=relevance_factor, top_n=limit
        )
        for i in reranked_indices:
            yield ranked_results[i]

    def matches(self, obj: OBJECT, **kwargs) -> Iterator[SEARCH_RESULT]:
        """
        Query the database for matches to an object.

        :param obj:
        :param kwargs:
        :return:
        """
        text_field = self.text_lookup
        text = self._text(obj, text_field)
        logger.info(f"Query term: {text}")
        yield from self.search(text, **kwargs)

    def lookup(self, id: str, collection: str = None, **kwargs) -> OBJECT:
        """
        Lookup an object by its ID.

        :param id:
        :param collection:
        :return:
        """
        client = self.client
        collection = client.get_or_create_collection(name=self._get_collection(collection))
        results = collection.get([id], include=["metadatas"])
        return self._unjson(results["metadatas"][0])

    def collections(self) -> Iterator[str]:
        """
        Return the names of all collections in the database.

        :return:
        """
        client = self.client
        for c in client.list_collections():
            yield c.name

    def _is_openai(self, collection: Collection):
        if collection.metadata.get("model", "").startswith("openai:"):
            return True

    def peek(self, collection: str = None, limit=5, offset: int = 0, **kwargs) -> Iterator[OBJECT]:
        c = self.client.get_collection(name=self._get_collection(collection))
        logger.debug(f"Peeking at {collection} with limit={limit}, offset={offset}")
        results = c.peek(limit=limit)
        logger.debug(f"Got {len(results)} results")
        # TODO: DRY
        metadatas = results["metadatas"]
        for i in range(0, len(metadatas)):
            yield self._unjson(metadatas[i])

    def fetch_all_objects_memory_safe(
        self, collection: str = None, batch_size: int = 100, **kwargs
    ) -> Iterator[OBJECT]:
        """
        Fetch all objects from a collection, in batches to avoid memory overload.
        """
        offset = 0
        client = self.client
        collection_obj = client.get_collection(name=self._get_collection(collection))
        while True:
            results = collection_obj.get(
                offset=offset,
                limit=batch_size,
                include=["metadatas", "embeddings", "documents"],
                **kwargs,
            )
            logger.info(f"Fetching batch from {offset}...")
            metadatas = results["metadatas"]
            documents = results["documents"]
            embeddings = results["embeddings"]
            if not documents and not metadatas and not embeddings:
                break
            for i in range(0, len(documents)):
                if not metadatas[i]:
                    logger.error(
                        f"Empty metadata for item {i} [num: {len(metadatas)}] doc: {documents[i]}"
                    )
                    continue
                obj = (
                    self._unjson(metadatas[i]),
                    0.0,
                    {
                        "document": documents[i],
                    },
                )
                if embeddings:
                    obj[2]["_embeddings"] = embeddings[i]
                yield obj
            offset += batch_size

    def dump_then_load(self, collection: str = None, target: DBAdapter = None):
        """
        Dump a collection to a file, then load it into another database.

        :param collection:
        :param target:
        :return:
        """
        client = self.client
        collection_obj = client.get_collection(name=self._get_collection(collection))
        if not isinstance(target, ChromaDBAdapter):
            raise ValueError("Target must be a ChromaDBAdapter")
        cm = self.collection_metadata(collection)
        ef = self._embedding_function(cm.model)
        # this currently prevents interadapter copying (duck to chroma)
        # target.get_collection (abstract) should be implemented
        target_collection_obj = target.client.get_or_create_collection(
            name=collection,
            embedding_function=ef,
            metadata=cm.dict(exclude_none=True),
        )
        result = collection_obj.get(include=["metadatas", "documents", "embeddings"])
        if not result["ids"]:
            raise ValueError("No ids found")
        if not result["embeddings"]:
            raise ValueError("No embeddings found")
        i = 0
        batch_size = 5000
        while i < len(result["ids"]):
            logger.debug(f"Dumping {i} of {len(result['ids'])}")
            batched_obj = {}
            for k in ["ids", "metadatas", "documents", "embeddings"]:
                batched_obj[k] = result[k][i : i + batch_size]
            target_collection_obj.add(**batched_obj)
            i += batch_size



"""Base Agent."""

from abc import ABC
from dataclasses import dataclass
from typing import Union

from curategpt import DBAdapter, Extractor
from curategpt.wrappers import BaseWrapper


@dataclass
class BaseAgent(ABC):  # noqa: B024
    """
    Base class for agents.

    An agent is capable of composing together different actions to achieve a goal.

    An agent typically has a *knowledge source* that is uses to search for information.
    An agent also has access to a model through an extractor.
    """

    knowledge_source: Union[DBAdapter, BaseWrapper] = None
    """A searchable source of information"""

    knowledge_source_collection: str = None

    extractor: Extractor = None
    """Engine performing LLM operations, including extracting from prompt responses"""

    def search(self):
        raise NotImplementedError("Search method must be implemented by subclass")



from .bioc_wrapper import BiocWrapper
from .pubmed_wrapper import PubmedWrapper
from .wikipedia_wrapper import WikipediaWrapper

__all__ = [
    "BiocWrapper",
    "PubmedWrapper",
    "WikipediaWrapper",
]



"""Extractor that uses OpenAI functions."""

import json
import logging
import os
from dataclasses import dataclass
from typing import List

from openai import OpenAI, OpenAIError

from curategpt.extract.extractor import AnnotatedObject, Extractor

FUNC_NAME = "extract_data"

logger = logging.getLogger(__name__)


@dataclass
class OpenAIExtractor(Extractor):
    """
    Extractor that uses OpenAI functions.
    """

    max_tokens: int = 3000
    model: str = "gpt-4"
    # conversation: List[Dict[str, Any]] = None
    # conversation_mode: bool = False

    @staticmethod
    def _get_openai_client():
        """
        Private method to get an instance of the OpenAI client.
        """
        api_key = os.getenv("OPENAI_API_KEY")
        if api_key is None:
            raise OpenAIError(
                "The api_key client option must be set either by passing api_key to the client or by setting the OPENAI_API_KEY environment variable"
            )
        return OpenAI(api_key=api_key)

    def functions(self):
        return [
            {
                "name": FUNC_NAME,
                "description": "A n ontology term",
                "parameters": self.schema_proxy.json_schema(),
            },
        ]

    def extract(
        self,
        text: str,
        target_class: str,
        examples: List[AnnotatedObject] = None,
        examples_as_functions=False,
        conversation=None,
        **kwargs,
    ) -> AnnotatedObject:
        messages = [
            {
                "role": "system",
                "content": f"You are system that returns {target_class} object in JSON.",
            },
        ]
        for example in examples:
            ex_text = example.text
            ex_object = example.object
            # print(f"EXAMPLE = {ex_text}")
            messages.append(
                {
                    "role": "user",
                    "content": f"make terms for {ex_text}",
                }
            )
            if not examples_as_functions:
                messages.append(
                    {
                        "role": "assistant",
                        "content": None,
                        "function_call": {
                            "name": FUNC_NAME,
                            "arguments": json.dumps(ex_object),
                        },
                    },
                )
            else:
                messages.append(
                    {
                        "role": "function",
                        "name": FUNC_NAME,
                        "content": json.dumps(ex_object),
                    },
                )
        if conversation:
            messages.extend(conversation)
        # content = f"make terms for {text}"
        content = text
        messages.append(
            {
                "role": "user",
                "content": content,
            }
        )
        # print(yaml.dump(messages))
        client = self._get_openai_client()

        response = client.chat.completions.create(
            model=self.model,
            functions=self.functions(),
            messages=messages,
            max_tokens=self.max_tokens,
        )
        logger.debug(f"RESPONSE = {response}")
        # print(response)
        choice = response.choices[0]
        message = choice.message
        if "function_call" not in message:
            if self.raise_error_if_unparsable:
                raise ValueError("No function call in response")
            r = "{}"
        else:
            r = message["function_call"]["arguments"]
        try:
            obj = json.loads(r)
            if conversation:
                conversation.append(messages[-1])
                conversation.append({"role": "function", "name": FUNC_NAME, "content": r})
        except json.decoder.JSONDecodeError as e:
            if self.raise_error_if_unparsable:
                raise e
            obj = {}
        return AnnotatedObject(object=obj)



"""
Formats data objects for presentation to humans and machine agents.

Note: this package is not in heavy use just now.

Currently we remove identifiers from data objects before caching,
and present the LLM with the same kind of view a human would see
when using a curation tool like Protege.

In future, this will be separated, and the store will house the
normalized representation, and this layer will handle
presenting data to the LLM (or human) with IDs changed to labels.

TODO: will this also handle the reverse?
"""



"""Autocomplete objects using RAG."""

import logging
from dataclasses import dataclass, field
from typing import Any, ClassVar, Dict, List, Optional, Union

from pydantic import BaseModel, ConfigDict

from curategpt.agents.base_agent import BaseAgent
from curategpt.agents.chat_agent import ChatResponse
from curategpt.extract import AnnotatedObject
from curategpt.store import DBAdapter

logger = logging.getLogger(__name__)

OBJECT = Dict[str, Any]


class PredictedFieldValue(BaseModel):

    model_config = ConfigDict(protected_namespaces=())
    id: str
    original_id: Optional[str] = None
    predicted_value: Optional[str] = None
    current_value: Optional[str] = None
    field_predicted: Optional[str] = None


def _dict2str(d: Dict[str, Any]) -> str:
    toks = []
    for k, v in d.items():
        if v:
            toks.append(f"{k}={v}")
    return ", ".join(toks)


@dataclass
class DatabaseAugmentedStructuredExtraction(BaseAgent):
    """
    Extracts structured objects from unstructured documents.

    This implements a standard knowledgebase retrieval augmented generation pattern;
    the knowledge_source is queried for relevant objects; these are presented
    as *examples* to a LLM query, via an extractor.
    """

    document_adapter: DBAdapter = None
    """Adapter to supplementary knowledge in unstructured form."""

    document_adapter_collection: str = None
    """Collection to use for document adapter.
    NOTE: may be deprecated as now collections can be bound to adapters
    """

    default_target_class: ClassVar[str] = "Thing"

    conversation: List[Dict[str, Any]] = None  # TODO
    conversation_mode: bool = False  # TODO
    relevance_factor: float = 0.5
    """Relevance factor for diversifying search results using MMR."""

    max_background_document_size: int = 1000
    """TODO: more sophisticated way to estimate size of background document."""

    background_document_limit: int = 3
    """Number of background documents to use. TODO: more sophisticated way to estimate."""

    default_masked_fields: List[str] = field(default_factory=lambda: ["original_id"])

    def extract(
        self,
        text: Union[str, ChatResponse],
        target_class: str = None,
        feature_fields: List[str] = None,
        generate_background=False,
        collection: str = None,
        rules: List[str] = None,
        fields_to_mask: List[str] = None,
        fields_to_predict: List[str] = None,
        merge=True,
        **kwargs,
    ) -> AnnotatedObject:
        """
        Populate structured object from text

        :param seed:
        :param target_class:
        :param context_property:
        :param generate_background:
        :param collection:
        :param rules:
        :param kwargs:
        :return:
        """
        if fields_to_mask is None:
            fields_to_mask = []
        extractor = self.extractor
        if not target_class:
            cm = self.knowledge_source.collection_metadata(collection)
            if not cm:
                raise ValueError(f"Invalid collection: {collection}")
            target_class = cm.object_type
        if not target_class:
            target_class = self.default_target_class

        annotated_examples = []
        logger.debug(f"Searching for seed: {text}")
        for obj, _, _obj_meta in self.knowledge_source.search(
            text,
            relevance_factor=self.relevance_factor,
            collection=collection,
            **kwargs,
        ):
            min_obj = {
                k: v
                for k, v in obj.items()
                if v
                and isinstance(v, str)
                and k not in fields_to_mask
                and (not fields_to_predict or k in fields_to_predict)
            }
            if not min_obj:
                continue
            ae = AnnotatedObject(object=obj)
            annotated_examples.append(ae)
        if not annotated_examples:
            raise ValueError(f"No examples found for seed: {text}")
        docs = []
        if self.document_adapter:
            logger.debug("Adding background knowledge.")
            for _obj, _, obj_meta in self.document_adapter.search(
                text,
                limit=self.background_document_limit,
                collection=self.document_adapter_collection,
            ):
                obj_text = obj_meta["document"]
                # TODO: use tiktoken to estimate
                obj_text = obj_text[0 : self.max_background_document_size]
                docs.append(obj_text)
        if generate_background:
            # prompt = f"Generate a comprehensive description about the {target_class} with {context_property} = {seed}"
            response = extractor.model.prompt(
                f"Describe the {target_class} in the following text: {text}"
            )
            if docs is None:
                docs = []
            docs.append(response.text())
        if docs:
            # TODO: experiment with placing after examples.
            background = "BACKGROUND:"
            background += "\n\n" + "\n\n".join(docs)
            background += "\n---\n"
            logger.info(f"Background: {background}")
        else:
            background = None
        ao = extractor.extract(
            f"Generate a {target_class} from the following text: {text}",
            target_class=target_class,
            examples=annotated_examples,
            background_text=background,
            rules=rules,
            min_examples=2,
        )
        return ao



# https://github.com/openai/openai-cookbook/blob/main/examples/How_to_count_tokens_with_tiktoken.ipynb
import logging
from typing import List, Optional

import tiktoken

logger = logging.getLogger(__name__)


def max_tokens_by_model(model_id: Optional[str] = None):
    """
    Return the maximum number of tokens allowed by a model.

    TODO: return precise values, currently an estimate.
    """
    if model_id == "gpt-4":
        return 8192
    elif "gpt-4o-mini" in model_id:
        return 16384
    elif "gpt-4o" in model_id:
        return 1024 * 128
    elif model_id == "gpt-3.5-turbo-16k":
        return 16384
    else:
        return 4097


def estimate_num_tokens(messages: List[str], model="gpt-4"):
    """
    Return the number of tokens used by a list of messages.

    Note: this is an estimate
    """
    try:
        encoding = tiktoken.encoding_for_model(model)
    except KeyError:
        logger.warning("Warning: model not found. Using cl100k_base encoding.")
        encoding = tiktoken.get_encoding("cl100k_base")
    if model in {
        "gpt-3.5-turbo-0613",
        "gpt-3.5-turbo-16k-0613",
        "gpt-4-0314",
        "gpt-4-32k-0314",
        "gpt-4-0613",
        "gpt-4-32k-0613",
    }:
        tokens_per_message = 3
        # tokens_per_name = 1
    elif model == "gpt-3.5-turbo-0301":
        tokens_per_message = 4  # every message follows <|start|>{role/name}\n{content}<|end|>\n
        # tokens_per_name = -1  # if there's a name, the role is omitted
    elif "gpt-3.5-turbo" in model:
        logger.info(
            "Warning: gpt-3.5-turbo may update over time. Returning num tokens assuming gpt-3.5-turbo-0613."
        )
        return estimate_num_tokens(messages, model="gpt-3.5-turbo-0613")
    elif "gpt-4" in model:
        logger.info(
            "Warning: gpt-4 may update over time. Returning num tokens assuming gpt-4-0613."
        )
        return estimate_num_tokens(messages, model="gpt-4-0613")
    else:
        raise NotImplementedError(
            f"num_tokens_from_messages() is not implemented for model {model}."
            "See https://github.com/openai/openai-python/blob/main/chatml.md"
            " for information on how messages are converted to tokens."
        )
    num_tokens = 0
    for message in messages:
        num_tokens += tokens_per_message
        num_tokens += len(encoding.encode(message))
    num_tokens += 3  # every reply is primed with <|start|>assistant<|message|>
    return num_tokens



from typing import Dict, Optional

from pydantic import BaseModel, ConfigDict


class CollectionMetadata(BaseModel):
    """
    Metadata about a collection.

    This is an open class, so additional metadata can be added.
    """

    model_config = ConfigDict(protected_namespaces=())

    name: Optional[str] = None
    """Name of the collection"""

    description: Optional[str] = None
    """Description of the collection"""

    model: Optional[str] = None
    """Name of any ML model"""

    object_type: Optional[str] = None
    """Type of object in the collection"""

    source: Optional[str] = None
    """Source of the collection"""

    # DEPRECATED
    annotations: Optional[Dict] = None
    """Additional metadata"""

    object_count: Optional[int] = None
    """Number of objects in the collection"""

    hnsw_space: Optional[str] = None
    """Space used for hnsw index (e.g. 'cosine')"""



from enum import Enum
from pathlib import Path
from typing import Dict, List, Optional, Union

from pydantic import BaseModel, ConfigDict


class StratifiedCollection(BaseModel):
    """
    A collection of objects that have been split into training, test, and validation sets.
    """

    model_config = ConfigDict(protected_namespaces=())
    source: str = None
    training_set: List[Dict] = None
    testing_set: List[Dict] = None
    validation_set: Optional[List[Dict]] = None
    training_set_collection: Optional[str] = None
    testing_set_collection: Optional[str] = None
    validation_set_collection: Optional[str] = None


class ClassificationOutcome(str, Enum):
    TRUE_POSITIVE = "True Positive"
    TRUE_NEGATIVE = "True Negative"
    FALSE_POSITIVE = "False Positive"
    FALSE_NEGATIVE = "False Negative"


class AggregationMethod(str, Enum):
    MACRO = "macro"
    MICRO = "micro"
    WEIGHTED = "weighted"


class ClassificationMetrics(BaseModel):

    model_config = ConfigDict(protected_namespaces=())
    precision: float
    recall: float
    f1_score: float
    accuracy: float
    specificity: float
    true_positives: int = (None,)
    true_negatives: int = (None,)
    false_positives: int = (None,)
    false_negatives: int = (None,)


class Task(BaseModel):
    """
    A task to be run by the evaluation runner.
    """

    model_config = ConfigDict(protected_namespaces=())
    model_name: str = "gpt-3.5-turbo"
    embedding_model_name: str = "openai:"
    generate_background: bool = False
    task_started: Optional[str] = None
    task_finished: Optional[str] = None
    executed_on: Optional[str] = None
    agent: Optional[str] = "dae"
    extractor: Optional[str] = "BasicExtractor"
    method: Optional[str] = None
    source_db_path: Optional[str] = None
    target_db_path: Optional[str] = None
    source_collection: Optional[str] = None
    additional_collections: Optional[List[str]] = None
    num_training: int = None
    num_testing: int = None
    num_validation: int = 0
    stratified_collection: Optional[StratifiedCollection] = None
    fields_to_mask: Optional[List[str]] = None
    fields_to_predict: Optional[List[str]] = None
    report_path: Optional[str] = None
    working_directory: Optional[Union[Path, str]] = None
    results: Optional[ClassificationMetrics] = None

    @property
    def id(self) -> str:
        pred = ".".join(self.fields_to_predict) if self.fields_to_predict else ""
        mask = ".".join(self.fields_to_mask) if self.fields_to_mask else ""
        model = self.model_name.replace(":", "")
        em_model = self.embedding_model_name.replace(":", "")
        id = (
            f"{self.source_collection}-P{pred}-M{mask}-BG{self.generate_background}-"
            f"Tr{self.num_training}-Te{self.num_testing}-M{model}-EM{em_model}"
        )
        if self.method:
            id += f"-Mth{self.method}"
        if self.additional_collections:
            id += f"-AC{'-'.join(self.additional_collections)}"
        return id



"""Chat with a KB."""

import logging
import time
from abc import ABC
from dataclasses import dataclass, field
from typing import ClassVar, Dict, List

import requests
import requests_cache
import xmltodict
from eutils import Client

from curategpt.wrappers import BaseWrapper

logger = logging.getLogger(__name__)

ESEARCH_URL = "https://eutils.ncbi.nlm.nih.gov/entrez/eutils/esearch.fcgi"
EFETCH_URL = "https://eutils.ncbi.nlm.nih.gov/entrez/eutils/{tool}.fcgi"


@dataclass
class EUtilsWrapper(BaseWrapper, ABC):
    """
    A wrapper to provide a search facade over NCBI Biosample.

    This is a dynamic wrapper: it can be used as a search facade,
    but cannot be ingested in whole.
    """

    name: ClassVar[str] = "__eutils__"

    eutils_db: ClassVar[str] = None

    id_prefix: ClassVar[str] = None

    fetch_tool: ClassVar[str] = "efetch"

    eutils_client: Client = None

    session: requests.Session = field(default_factory=lambda: requests.Session())

    _uses_cache: bool = False

    def set_cache(self, name: str) -> None:
        self.session = requests_cache.CachedSession(name)
        self._uses_cache = True

    def external_search(self, text: str, expand: bool = True, **kwargs) -> List[Dict]:
        db = self.eutils_db
        if expand:
            logger.info(f"Expanding search term: {text} to create {db} query")
            model = self.extractor.model
            response = model.prompt(
                text, system="generate a semi-colon separated list of the most relevant terms"
            )
            terms = response.text().split(";")
            search_term = " OR ".join(terms)
        else:
            search_term = text
        logger.info(f"Constructed search term: {search_term}")
        # Parameters for the request
        params = {
            "db": db,
            "term": search_term,
            "retmax": 100,
            "sort": "relevance",
            "retmode": "json",
        }

        time.sleep(0.5)
        response = requests.get(ESEARCH_URL, params=params)
        data = response.json()

        # Extract IDs from the response
        ids = data["esearchresult"]["idlist"]

        if not ids:
            logger.warning(f"No results with {search_term}")
            if expand:
                logger.info("Trying again without expansion")
                return self.external_search(text, expand=False, **kwargs)
            else:
                logger.error(f"Failed to find results for {text}")
                return []

        logger.info(f"Found {len(ids)} results: {ids}")
        return self.objects_by_ids(ids)

    def objects_by_ids(self, object_ids: List[str]) -> List[Dict]:
        id_prefix = self.id_prefix or self.eutils_db
        object_ids = sorted([x.replace(f"{id_prefix}:", "") for x in object_ids])
        session = self.session
        logger.debug(f"Using session: {session} [cached: {self._uses_cache} for {object_ids}")

        # Parameters for the efetch request
        efetch_params = {
            "db": self.eutils_db,
            "id": ",".join(object_ids),  # Combine  IDs into a comma-separated string
            "retmode": "xml",
        }
        # if not self._uses_cache:
        time.sleep(0.25)
        tool = self.fetch_tool
        efetch_response = session.get(EFETCH_URL.format(tool=tool), params=efetch_params)
        if not efetch_response.ok:
            logger.error(f"Failed to fetch data for {object_ids}")
            raise ValueError(
                f"Failed to fetch data for {object_ids} using {session} and {efetch_params}"
            )
        results = xmltodict.parse(efetch_response.text)
        return self.objects_from_dict(results)

    def objects_from_dict(self, results: Dict) -> List[Dict]:
        raise NotImplementedError



"""
CurateGPT Agents.

These chain together different search and generate components.
"""

from .chat_agent import ChatAgent
from .dragon_agent import DragonAgent
from .evidence_agent import EvidenceAgent
from .mapping_agent import MappingAgent

__all__ = ["MappingAgent", "ChatAgent", "DragonAgent", "EvidenceAgent"]



"""Retrieval Augmented Generation (RAG) Base Class."""

import logging
from dataclasses import dataclass, field
from typing import Any, ClassVar, Dict, Iterable, List, Optional, Tuple, Union

import yaml
from pydantic import BaseModel, ConfigDict

from curategpt.agents.base_agent import BaseAgent
from curategpt.extract import AnnotatedObject
from curategpt.store import DBAdapter

logger = logging.getLogger(__name__)

OBJECT = Dict[str, Any]


class PredictedFieldValue(BaseModel):

    model_config = ConfigDict(protected_namespaces=())
    id: str
    original_id: Optional[str] = None
    predicted_value: Optional[str] = None
    current_value: Optional[str] = None
    field_predicted: Optional[str] = None


def _dict2str(d: Dict[str, Any]) -> str:
    toks = []
    for k, v in d.items():
        if v:
            toks.append(f"{k}={v}")
    return ", ".join(toks)


@dataclass
class DragonAgent(BaseAgent):
    """
    Retrieves objects in response to a query using a structured knowledge source.

    (essentially a structured object autocomplete)

    This implements a standard knowledgebase retrieval augmented generation pattern;
    the knowledge_source is queried for relevant objects; these are presented
    as *examples* to a LLM query, via an extractor.
    """

    document_adapter: DBAdapter = None
    """Adapter to supplementary knowledge in unstructured form."""

    document_adapter_collection: str = None
    """Collection to use for document adapter.
    NOTE: may be deprecated as now collections can be bound to adapters
    """

    default_target_class: ClassVar[str] = "Thing"

    conversation: List[Dict[str, Any]] = None  # TODO
    conversation_mode: bool = False  # TODO
    relevance_factor: float = 0.5
    """Relevance factor for diversifying search results using MMR."""

    max_background_document_size: int = 2000
    """TODO: more sophisticated way to estimate size of background document."""

    background_document_limit: int = 3
    """Number of background documents to use. TODO: more sophisticated way to estimate."""

    default_masked_fields: List[str] = field(default_factory=lambda: ["original_id"])

    def complete(
        self,
        seed: Union[str, Dict[str, Any]],
        target_class: str = None,
        context_property: str = None,
        generate_background=False,
        collection: str = None,
        rules: List[str] = None,
        fields_to_mask: List[str] = None,
        fields_to_predict: List[str] = None,
        merge=True,
        **kwargs,
    ) -> AnnotatedObject:
        """
        Populate structured object from partially populated object.

        If a string is passed, then an object of form ``{context_property: seed}`` is used.

        :param seed:
        :param target_class:
        :param context_property:
        :param generate_background:
        :param collection:
        :param rules: these are included in the prompt
        :param kwargs:
        :return:
        """
        extractor = self.extractor
        if not target_class:
            cm = self.knowledge_source.collection_metadata(collection)
            if not cm:
                raise ValueError(f"Invalid collection: {collection}")
            target_class = cm.object_type
        if not target_class:
            target_class = self.default_target_class
        # construct partially populated seed object
        if context_property is None:
            context_property = "label"
        if isinstance(seed, str):
            seed = {context_property: seed}
        if isinstance(seed, dict):
            feature_fields = [k for k in seed.keys() if seed[k]]
        else:
            if context_property:
                feature_fields = [context_property]
            else:
                feature_fields = []
        if fields_to_mask is None:
            fields_to_mask = self.default_masked_fields

        def generate_input_str(obj: Union[str, Dict], prefix="Structured representation of") -> str:
            if isinstance(obj, dict):
                if not feature_fields:
                    # generate a complete object
                    min_obj = {
                        k: v
                        for k, v in obj.items()
                        if v and isinstance(v, str) and k not in fields_to_mask
                    }
                else:
                    # generate a partial object with only feature fields set
                    min_obj = {k: obj[k] for k in feature_fields if k in obj}
                if min_obj:
                    as_text = yaml.safe_dump(min_obj, sort_keys=False).strip()
                    return f"{prefix} {target_class} with {as_text}"
                else:
                    return f"{prefix} {target_class}:"
            elif isinstance(obj, str):
                return f"{prefix} {target_class} with {context_property} = {obj}"
            else:
                raise ValueError(f"Invalid type for obj: {type(obj)} //  {obj}")

        annotated_examples = []
        seed_search_term = seed if isinstance(seed, str) else yaml.safe_dump(seed, sort_keys=True)
        logger.debug(f"Searching for seed: {seed_search_term}")
        for obj, _, _obj_meta in self.knowledge_source.search(
            seed_search_term,
            relevance_factor=self.relevance_factor,
            collection=collection,
            **kwargs,
        ):
            # training example input
            input_text = generate_input_str(obj)
            # training example output
            obj_predicted_part = {
                k: v
                for k, v in obj.items()
                if v
                and k not in feature_fields
                and (not fields_to_predict or k in fields_to_predict)
            }
            if not obj_predicted_part:
                logger.warning(
                    f"Skipping; Candidate example lacked predictable properties ({fields_to_predict}): {obj}; "
                    f"Context properties: {feature_fields}; "
                    f"Num examples={len(annotated_examples)}"
                )
                continue
            ae = AnnotatedObject(object=obj_predicted_part, annotations={"text": input_text})
            annotated_examples.append(ae)
        if not annotated_examples:
            logger.error(f"No suitable examples found for seed: {seed}")
        docs = []
        if self.document_adapter:
            logger.debug("Adding background knowledge.")
            for _obj, _, obj_meta in self.document_adapter.search(
                seed_search_term,
                limit=self.background_document_limit,
                collection=self.document_adapter_collection,
            ):
                obj_text = obj_meta["document"]
                # TODO: use tiktoken to estimate
                obj_text = obj_text[0 : self.max_background_document_size]
                docs.append(obj_text)
        gen_text = generate_input_str(seed)
        if generate_background:
            # prompt = f"Generate a comprehensive description about the {target_class} with {context_property} = {seed}"
            prompt = generate_input_str(
                seed, prefix="Generate a comprehensive description about the"
            )
            response = extractor.model.prompt(prompt)
            if docs is None:
                docs = []
            docs.append(response.text())
        if docs:
            # TODO: experiment with placing after examples.
            background = "BACKGROUND:"
            background += "\n\n" + "\n\n".join(docs)
            background += "\n---\n"
            logger.info(f"Background: {background}")
        else:
            background = None
        ao = extractor.extract(
            gen_text,
            target_class=target_class,
            examples=annotated_examples,
            background_text=background,
            rules=rules,
        )
        if merge:
            # put the seed object (non-predicted fields) back in
            init_object = {}
            for k, v in seed.items():
                if v and not ao.object.get(k, None):
                    init_object[k] = v
            ao.object = {**init_object, **ao.object}
        return ao

    def generate_all(
        self,
        collection: str,
        field_to_predict: str,
        missing_only=True,
        object_ids: Optional[Iterable[str]] = None,
        **kwargs,
    ) -> Iterable[Tuple[str, str, Any, Any]]:
        """
        Generate missing value for a field for all objects in a collection.

        :param collection:
        :param field_to_predict:
        :param missing_only:
        :param object_ids:
        :param kwargs:
        :return:
        """
        logger.info("Fetching all objects...")
        it = self.knowledge_source.find(
            {},
            collection=collection,
            limit=100000,
        )
        for obj, _, __ in it:
            obj_id = obj["id"]
            original_id = obj.get("original_id", None)
            logger.debug(f"Checking {obj_id} // {object_ids}")
            if object_ids and obj_id not in object_ids and original_id not in object_ids:
                continue
            curr_val = obj.get(field_to_predict, None)
            if missing_only and curr_val:
                logger.debug(f"Skipping; {field_to_predict} already present: {curr_val}")
                continue
            ao = self.complete(obj, collection=collection, **kwargs)
            yield PredictedFieldValue(
                id=obj_id,
                original_id=original_id,
                field_predicted=field_to_predict,
                predicted_value=ao.object.get(field_to_predict, None),
                current_value=curr_val,
            )

    def generate_queries(self, context_property="name", n=5, **kwargs) -> List[str]:
        ks = self.knowledge_source
        ids = []
        for obj, _, _x in ks.find({}, **kwargs):
            ids.append(obj.get(context_property))
        system = f"""Suggest {n} additional entities to complement the existing supplied entities.
        Return as a comma separated list. Do not return additional information. Keep the same
        level of specificity. You can return similar entities, but not identical/equivalent.

        ## Example (n=2)

        Current:
        Margarita Pizza, Hawaiian Pizza, Veggie Pizza

        Suggested:
        Mushroom Pizza, Meat Lovers Pizza.
        """
        ids_str = ", ".join(ids)
        prompt = f"""Current:
        {ids_str}
        Suggested:
        """
        logger.info(f"PROMPT: {prompt}")
        response = self.extractor.model.prompt(prompt, system=system)
        txt = response.text()
        if "." in txt:
            txt = txt[0 : txt.index(".")]
        suggestions = txt.split(",")
        suggestions = [x.strip() for x in suggestions]
        ids_norm = [x.lower() for x in ids]
        suggestions = [x for x in suggestions if x.lower() not in ids_norm]
        return suggestions

    def review(
        self,
        obj: dict,
        context_property: str = None,
        rules=None,
        collection=None,
        fields_to_predict=None,
        primary_key=None,
        **kwargs,
    ) -> AnnotatedObject:
        """
        Review an object for correctness, completeness, and consistency.

        :param obj:
        """
        if fields_to_predict and not context_property:
            raise ValueError("context_property is required if fields_to_predict")

        pk_val = obj.get(primary_key, None)

        def _obj_as_str(obj: dict) -> str:
            slim_obj = {
                k: v
                for k, v in obj.items()
                if v is not None
                and ((not fields_to_predict or k in fields_to_predict) or k == context_property)
            }
            return yaml.dump(slim_obj, sort_keys=True)

        q = yaml.safe_dump(obj, sort_keys=True)
        texts = []
        for example_obj, _, _obj_meta in self.knowledge_source.search(
            q,
            relevance_factor=self.relevance_factor,
            collection=collection,
            **kwargs,
        ):
            if primary_key:
                if example_obj.get(primary_key, None) == pk_val:
                    logger.debug(
                        f"Skipping example with same primary key: {primary_key} = {pk_val}"
                    )
                    continue
            texts.append(_obj_as_str(example_obj))
        system = """
        Your are an expert database curator. Your job is to take an input record and review it for correctness,
        completeness, and consistency. In particular, the input record may be incomplete so you should
        use your expert domain knowledge to fill things in more detail and granularity,

        I will give you a list of example objects that are similar to the
        input record for you to follow as a guide to the schema (but these may themselves be incomplete).

        You should return results in YAML in a ```yaml...``` block.
        You should follow the examples for a consistent schema, but you can add an optional
        `review_notes:` field at either the top level or within individual objects to provide additional
        commentary.
        """
        if rules:
            system += f"\n\nAdditional considerations: {'. '.join(list(rules))}"
        prompt = "Example records:\n\n" + "\n\n".join(texts)

        prompt += f"\n\nInput record:\n\n{_obj_as_str(obj)}"
        response = self.extractor.model.prompt(prompt, system=system)
        ao = self.extractor.deserialize(response.text(), format="yaml")
        if not isinstance(ao.object, dict):
            logger.warning(f"Expected dict, got {ao.object}")
            if isinstance(ao.object, list):
                logger.warning(f"Taking first element of list of len {len(ao.object)}")
                ao.object = ao.object[0]
        if isinstance(ao.object, dict):
            if context_property:
                ao.object[context_property] = obj[context_property]
            if fields_to_predict:
                for k, v in obj.items():
                    if k not in fields_to_predict:
                        ao.object[k] = v
        ao.annotations["prompt"] = prompt
        return ao



from dataclasses import dataclass
from typing import Dict

import yaml
from jinja2 import Template
from pydantic import BaseModel, ConfigDict

from curategpt.agents.base_agent import BaseAgent
from curategpt.conf.prompts import PROMPTS_DIR
from curategpt.extract import AnnotatedObject


class KnowledgeBaseSpecification(BaseModel):
    model_config = ConfigDict(protected_namespaces=())
    kb_name: str
    description: str
    attributes: str
    main_class: str


@dataclass
class BootstrapAgent(BaseAgent):

    def bootstrap_schema(self, specification: KnowledgeBaseSpecification) -> AnnotatedObject:
        """
        Bootstrap a schema for a knowledge base.

        :param specification: Specification for the knowledge base.
        :return:
        """
        # use a jinja template
        template = Template(open(PROMPTS_DIR / "bootstrap-schema.j2").read())
        prompt = template.render(**specification.model_dump())
        extractor = self.extractor
        model = extractor.model
        response = model.prompt(prompt)
        ao = extractor.deserialize(response.text(), format="yaml")
        return ao

    def bootstrap_data(
        self, specification: KnowledgeBaseSpecification = None, schema: Dict = None
    ) -> str:
        """
        Bootstrap data for a knowledge base.

        :param specification: Specification for the knowledge base.
        :param schema: Schema for the knowledge base.
        :return:
        """
        spec_str = None
        if specification is not None:
            preamble = "Use the following knowledge base schema specification"
            spec_str = yaml.dump(specification.model_dump(), sort_keys=False)
        if schema is not None:
            preamble = (
                "Make sure the object you create is consistent with the following linkml schema"
            )
            spec_str = yaml.dump(schema, sort_keys=False)
        if not spec_str:
            raise ValueError("No specification or schema provided")
        extractor = self.extractor
        model = extractor.model
        prompt = (
            "Generate an example instance data objects in YAML, inside a ```yaml...``` block\n"
            "Choose any objects you like, but make them representative of the domain, "
            "with rich comprehensive information\n"
            "Separate each element with '---' breaks, such that each entry is a separate object\n"
            "Do not include the root container class"
        )
        prompt += f"{preamble}:\n{spec_str}"
        response = model.prompt(prompt)
        txt = response.text()
        if "```" in txt:
            txt = txt.split("```")[1]
            txt = txt.strip()
            if txt.startswith("yaml"):
                txt = txt[4:]
                txt = txt.strip()
        return txt



"""Chat with a KB."""

import logging
from dataclasses import dataclass, field
from time import sleep
from typing import ClassVar, Dict, Iterable, Iterator, Optional

import requests_cache

from curategpt.wrappers import BaseWrapper

URL = "https://fusion.ess-dive.lbl.gov/deepdive?rowStart={cursor}&pageSize={limit}"


def _get_records_chunk(session: requests_cache.CachedSession, cursor=1, limit=200) -> dict:
    """
    Get a chunk of records from ESSDeepDive.

    :param cursor:
    :return:
    """
    url = URL.format(limit=limit, cursor=cursor)
    response = session.get(url)
    if response.status_code == 200:
        return response.json()
    else:
        raise ValueError(f"Could not download records from {url}")


def get_records(
    session: requests_cache.CachedSession, cursor=1, limit=200, maximum: int = None
) -> Iterator[dict]:
    """
    Iterate through all records in ESSDeepDive and download them.

    :param cursor:
    :return:
    """
    # note: do NOT do this recursively, as it will cause a stack overflow
    initial_chunk = _get_records_chunk(session, cursor=cursor, limit=limit)
    yield from initial_chunk["results"]
    while True:
        cursor += 1
        next_chunk = _get_records_chunk(session, cursor=cursor, limit=limit)
        yield from next_chunk["results"]
        pc = next_chunk["pageCount"]
        logger.warning(f"Got {pc} pages")
        if not pc:
            break
        sleep(0.1)


logger = logging.getLogger(__name__)


@dataclass
class ESSDeepDiveWrapper(BaseWrapper):
    """
    A wrapper over the ESSDeepDive API.
    """

    name: ClassVar[str] = "ess_deepdive"

    default_object_type = "Class"

    session: requests_cache.CachedSession = field(
        default_factory=lambda: requests_cache.CachedSession("ess_deepdive")
    )

    limit: int = field(default=50)

    def objects(
        self, collection: str = None, object_ids: Optional[Iterable[str]] = None, **kwargs
    ) -> Iterator[Dict]:
        seen = set()
        for record in get_records(self.session, limit=self.limit):
            id = record["data_file_url"] + "-" + record["field_name"]
            if id in seen:
                logger.warning(f"Skipping duplicate record {id}")
                continue
            seen.add(id)
            record["id"] = id
            yield record



"""Chat with a KB."""

import logging
from dataclasses import dataclass
from typing import ClassVar, Dict, Iterable, Iterator, Optional

import requests
from oaklib import BasicOntologyInterface, get_adapter

from curategpt.wrappers import BaseWrapper

URL = "https://api.microbiomedata.org/biosamples?per_page={limit}&page={cursor}"


def _get_samples_chunk(cursor=1, limit=200) -> dict:
    """
    Get a chunk of samples from NMDC.

    :param cursor:
    :return:
    """
    url = URL.format(limit=limit, cursor=cursor)
    response = requests.get(url)
    if response.status_code == 200:
        return response.json()
    else:
        raise ValueError(f"Could not download samples from {url}")


def get_samples(cursor=1, limit=200, maximum: int = None) -> Iterator[dict]:
    """
    Iterate through all samples in NMDC and download them.

    :param cursor:
    :return:
    """
    # note: do NOT do this recursively, as it will cause a stack overflow
    initial_chunk = _get_samples_chunk(cursor=cursor, limit=limit)
    num = initial_chunk["meta"]["count"]
    if maximum is not None:
        num = min(num, maximum)
    yield from initial_chunk["results"]
    while True:
        cursor += 1
        if cursor * limit >= num:
            break
        next_chunk = _get_samples_chunk(cursor=cursor, limit=limit)
        yield from next_chunk["results"]


logger = logging.getLogger(__name__)


@dataclass
class NMDCWrapper(BaseWrapper):
    """
    A wrapper over the NMDC Biosample API.
    """

    name: ClassVar[str] = "nmdc"

    default_object_type = "BioSample"

    _label_adapter: BasicOntologyInterface = None

    def objects(
        self, collection: str = None, object_ids: Optional[Iterable[str]] = None, **kwargs
    ) -> Iterator[Dict]:
        for sample in get_samples():
            for k in ["env_broad_scale", "env_local_scale", "env_medium"]:
                if k in sample:
                    self._labelify(sample[k]["term"])
            yield sample

    @property
    def label_adapter(self) -> BasicOntologyInterface:
        """Get the label adapter."""
        if self._label_adapter is None:
            self._label_adapter = get_adapter("sqlite:obo:envo")
        return self._label_adapter

    def _labelify(self, term: Dict):
        if "label" not in term or not term["label"]:
            term["label"] = self.label_adapter.label(term["id"])



from pathlib import Path

import yaml
from pydantic import BaseModel, ConfigDict


class DBSettings(BaseModel):

    model_config = ConfigDict(protected_namespaces=())

    name: str = "duckdb"
    """Name of the database."""

    model: str = "all-MiniLM-L6-v2"
    """Name of any embedding model"""

    hnsw_space: str = "cosine"
    """Space used for hnsw index (e.g. 'cosine')."""

    ef_construction: int = 128
    """
    Construction parameter for hnsw index.
    Higher values are more accurate but slower.
    """

    ef_search: int = 64
    """
    Search parameter for hnsw index.
    Higher values are more accurate but slower.
    """

    M: int = 16
    """M parameter for hnsw index"""

    def load_config(self, path: Path):
        with open(path) as file:
            config = yaml.safe_load(file)
            self.name = config.get("name", self.name)
            self.hnsw_space = config.get("hnsw_space", self.hnsw_space)
            self.model = config.get("model", self.model)
            self.ef_construction = config.get("ef_construction", self.ef_construction)
            self.ef_search = config.get("ef_search", self.ef_search)
            self.M = config.get("M", self.M)



from dataclasses import dataclass
from typing import Any, Iterable, Iterator, List, Tuple

from curategpt import DBAdapter


@dataclass
class Namer:
    """
    Maps between opaque IDs and names

    TODO: use caching
    """

    stores: List[DBAdapter]

    def map_object(self, obj: Any, paths: List[List[str]], forward=True) -> Any:
        """
        Maps an object to a new object with names instead of IDs
        """
        if isinstance(obj, dict):
            mapped = {}
            for k, v in obj.items():
                if [k] in paths:
                    mapped[k] = self.map_reference(v, forward=forward)
                else:
                    new_paths = [path[1:] for path in paths if path[0] == k]
                    mapped[k] = self.map_object(v, new_paths, forward=forward)
        elif isinstance(obj, list):
            mapped = []
            for v in obj:
                mapped.append(self.map_object(v, paths, forward=forward))
        else:
            mapped = obj
        return mapped

    def map_reference(self, ref: str, forward=True) -> str:
        """
        Maps a reference to a name
        """
        if forward:
            return list(self.ids_to_names([ref]))[0][1]
        else:
            return list(self.names_to_ids([ref]))[0][1]

    def ids_to_names(self, ids: Iterable[str]) -> Iterator[Tuple[str, str]]:
        """
        fetches mappings
        """
        yield from self._query_mappings(
            ids, self.store.identifier_field(), self.store.label_field()
        )

    def names_to_ids(self, names: Iterable[str]) -> Iterator[Tuple[str, str]]:
        """
        fetches mappings
        """
        yield from self._query_mappings(
            names, self.store.label_field(), self.store.identifier_field()
        )

    def _query_mappings(
        self, from_vals: List[str], from_field: str, to_field: str
    ) -> Iterator[Tuple[str, str]]:
        """
        fetches mappings
        """
        from_vals = list(from_vals)
        if not from_vals:
            return
        for store in self.stores:
            ors = [{from_field: from_val} for from_val in from_vals]
            if len(ors) == 1:
                where = ors[0]
            else:
                where = {"$or": ors}
            for obj, _, __ in store.find(where=where):
                from_val = obj.get(from_field, None)
                to_val = obj.get(to_field, None)
                if to_val:
                    yield from_val, to_val
                    from_vals.pop(from_val)
